:- use_module('Indexing/substitutiontree').
:- use_module(library(ordsets)).
:- use_module(library(terms)).
:- use_module(library(lists)).
:- compile(sets).
:- compile(subsumption).
:- compile(lpo).

:- dynamic testseries/1.
:- dynamic option/2.
:- dynamic modalLogic/3.
:- dynamic skolemFunction/4.
:- op(500,fx,not).

%---------------------------------------------------------------------------
% testSAT(temporalResolution,+VN,+AN,+ON,+RANDOM)
% applies the temporal resolution calculus to the random SNF formula identified
% by the parameters VN (number of propositions), AN (number of conjuncts), 
% ON (number of disjuncts), and RANDOM (number identifying the particular 
% formula).

testSAT(temporalResolution,VN,AN,ON,Random) :-
	generateFilename('.prolog.concept',VN,0,AN,ON,0,Random,FileName),
	format('TRP solving random PLTL formula ~d-0-~d-~d-0-~d~n',[VN,AN,ON,Random]),
	testFile(FileName).

%---------------------------------------------------------------------------
% testFile(+FileName)
% applies the temporal resolution calculus to the conjunction of SNF clauses
% contained in the file specified by FileName.
% Displays some statistics at the end, i.e. whether the set of SNF clauses is
% satisfiable, how long it took to decide it, etc.

testFile(FileName) :-
	see(FileName),
	read(and(SNFClauses)),
	seen,
	retractall(gensym_counter(_,_)),
	translateSNFClauses(SNFClauses,FOLClausesSNF),
	generateModalAxioms(FOLClausesModal),
	append(FOLClausesSNF,FOLClausesModal,FOLClauses),
	ifOption(debugLevel,DL,(DL>1),(write('Input clause set:'), nl,
				      printClauses(spass,debug,FOLClauses), nl)),
	retractall(skolemFunction(_,_,_)),
	!,
	sat(FOLClauses,_,CL),
	ifOption(debugLevel,DL,(DL>1),(write('Final clause set:'), nl,
				      printClauses(spass,debug,CL), nl)).

%---------------------------------------------------------------------------
% testFormula(+Formula)
% applies the temporal resolution calculus to the conjunction of SNF clauses
% Formula it is given as an argument.
% Displays some statistics at the end, i.e. whether the set of SNF clauses is
% satisfiable, how long it took to decide it, etc.

testFormula(Formula) :-
	Formula = and(SNFClauses),
	retractall(gensym_counter(_,_)),
	translateSNFClauses(SNFClauses,FOLClausesSNF),
	generateModalAxioms(FOLClausesModal),
	append(FOLClausesSNF,FOLClausesModal,FOLClauses),
	ifOption(debugLevel,DL,(DL>1),(write('Input clause set:'), nl,
				      printClauses(spass,debug,FOLClauses), nl)),
	retractall(skolemFunction(_,_,_)),
	!,
	sat(FOLClauses,_,CL),
	ifOption(debugLevel,DL,(DL>1),(write('Final clause set:'), nl,
				      printClauses(spass,debug,CL), nl)).

%---------------------------------------------------------------------------
% sat(+CL1,-Consistent,-CL2)
% applies the temporal resolution calculus to the list of SNF clauses CL1.
% Returns a saturated list of SNF clauses in the variable CL2 and the result
% of the satisfiability test in Consistent: If CL2 (and therefore CL1) is
% consistent, then Consistent will be `true', otherwise it will be `false'.
% Displays some statistics at the end, i.e. whether the set of SNF clauses is
% satisfiable, how long it took to decide it, etc.

sat(CL1,Consistent,CL) :-
        getRuntime(T0),
	setOption(mode,satisfy),
	getSometimeClauses(CL1,CLSometime,CLOther),
	generateAugmentation(CLSometime,CLAug),
	append(CLOther,CLAug,CLAll),
	computeCharacteristic(CLAll,CL2),
        length(CL2,N0),
	st_empty_tree(STEP),
	st_empty_tree(STEN),
	!,
	mainProofProcedure([[STEP,STEN],[]],CL2,CLSometime,Consistent,CL,statistics(NF,NR,NS,NTRSuccess,NTRFail,NTRNR,NC)),
	getRuntime(T1),
        T is T1 - T0,
	(Consistent == true -> 
	    format('Clause set is satisfiable~n',[])
	;
	    format('Clause set is unsatisfiable~n',[])
	),
        format('Total runtime:                                              ~t~3d~45| sec.~n',T),
        format('Number of clauses at the beginning:                         ~t~d~41|~n',N0),
        format('Number of clauses at the end:                               ~t~d~41|~n',NC),
	format('Number of initial factors    and step factors:              ~t~d~41|~n',NF),
	format('Number of initial resolvents and step resolvents:           ~t~d~41|~n',NR),
	format('Number of successful   applications of temporal resolution: ~t~d~41|~n',NTRSuccess),
	format('Number of nonredundant applications of temporal resolution: ~t~d~41|~n',NTRNR),
	format('Number of failed       applications of temporal resolution: ~t~d~41|~n',NTRFail),
	format('Number of subsumed clauses:                                 ~t~d~41|~n',NS),
	!.

%---------------------------------------------------------------------------
% generateAugmentation(+SOMETIMECLAUSES,-AUGMENTATIONCLAUSES)
% generates all the clauses derived by an application of temporal resolution
% to one of the clauses in SOMETIMECLAUSES, that are independent of the 
% loop clauses involved in an application of temporal resolution.

generateAugmentation([],[]) :-
	!.
generateAugmentation([cl([sometime(L)],NegLiterals,_PT)|CL0],CLAug) :-
	!,
	generateAugmentation(CL0,CLAug0),
	modNegateLiterals(NegLiterals,NegatedLiterals),
	term_variables(NegatedLiterals,[X]),
	term_variables(L,[X]),
	genWaitingPred(L,A1),
	Waiting =.. [A1,X],
	Clause1 = cl([Waiting,L|NegatedLiterals]),
	modExtractPredicate(L,[P]),
	NextP =.. [P,f(X)],
	(L = not(_) ->
	    NextL = not(NextP)
	; 
	    NextL = NextP
	),
	NextWaiting =.. [A1,f(X)],
	Clause2 = cl([not(Waiting),NextL,NextWaiting]),
	% test uh 07.06.2001
%       modNegateLiteral(L,NotL),
%	Clause3 = cl([not(Waiting),NotL]),
%	append([Clause1,Clause2,Clause3],CLAug0,CLAug).
	append([Clause1,Clause2],CLAug0,CLAug).
generateAugmentation([cl([sometime(L)],NegLiterals)|CL0],CLAug) :-
	generateAugmentation([cl([sometime(L)],NegLiterals,c0)|CL0],CLAug).
	
%---------------------------------------------------------------------------
% translateSNFCLAUSES(+SNFCLAUSELIST,-FOLCLAUSES)
% translates the list of SNF clauses SNFCLAUSELIST into a set of
% `pseudo' first-order clauses FOLCLAUSES.

translateSNFClauses([],[]) :-
	!.
translateSNFClauses([C1|CL],NCAll) :-
	translateSNFClause(C1,NC1),
	translateSNFClauses(CL,NCL),
	(functor(NC1,cl,_) ->
	    NCAll = [NC1|NCL]
	;
	    append(NC1,NCL,NCAll)
	),
	!.

%---------------------------------------------------------------------------
% translateSNFClause(+SNFCLAUSE,-FOLCLAUSES)
% translates the SNF clause SNFCLAUSE into a list of `pseudo' first-order 
% clauses.
% Basically,
% L1 | ... | Ln            becomes L1(now) | ... | Ln(now)  [for a constant now]
% G( L1 | ... | Ln )       becomes  L1(X) | ... | Ln(X) [for a variable X]
% G( L1 | ... | Ln | X(M1) | ... | X(Mk)) becomes
%    L1(X) | ... | Ln(X) | M1(f(X)) | ... Mk(f(X)) [for a fixed function symbol f]
% G( L1 | ... | Ln | F(L)) becomes L1(X) | ... | Ln(X) | sometime(L(X))
% and 
% G( L1 | ... | Ln | (-)box(M,A,L)) will be replaced by a set of clauses
% as described in our FroCoS paper.

translateSNFClause(always(or(LL)),NC) :-
	member(next(_),LL),
	!,
	% next SNF clause
	snfClassifyLiterals(LL,PropositionalLiterals,NextLiterals),
	convertLiterals(PropositionalLiterals,X,FOLLiterals0),
	convertLiterals(NextLiterals,X,FOLLiterals1),
	liftLiterals(FOLLiterals1,X,FOLLiterals2),
	append(FOLLiterals0,FOLLiterals2,FOLLiterals),
	splitLiterals(FOLLiterals,Pos,Neg),
	gensym(n,PT),
	NC = cl(Pos,Neg,PT).
translateSNFClause(always(or(LL)),NC) :-
	member(sometime(_),LL),
	!,
	% sometime SNF clause
	snfClassifyLiterals(LL,PropositionalLiterals,SometimeLiterals),
	convertLiterals(PropositionalLiterals,X,FOLLiterals0),
	convertLiterals(SometimeLiterals,X,FOLLiterals1),
	putBackOperator(FOLLiterals1,sometime,FOLLiterals2),
	append(FOLLiterals0,FOLLiterals2,FOLLiterals),
	splitLiterals(FOLLiterals,Pos,Neg),
	gensym(s,PT),
	NC = cl(Pos,Neg,PT).
translateSNFClause(always(or(LL)),[NC|Clauses]) :-
	(member(box(M,A,L),LL) ; member(not(box(M,A,L)),LL)),
	 !,
	 snfClassifyLiterals(LL,PropositionalLiterals,ModalLiterals),
	 convertLiterals(PropositionalLiterals,X,FOLLiterals0),
	 shallowModalTranslation(ModalLiterals,X,FOLLiterals1),
	 append(FOLLiterals0,FOLLiterals1,FOLLiterals),
	 splitLiterals(FOLLiterals,Pos,Neg),
	 gensym(m,PT),
	 NC = cl(Pos,Neg,PT),
	 deepModalTranslation(ModalLiterals,FOLLiterals1,X,Clauses).
translateSNFClause(always(or(LL)),NC) :-
	!,
	% global SNF clause
	convertLiterals(LL,_X,FOLLiterals),
	splitLiterals(FOLLiterals,Pos,Neg),
	gensym(g,PT),
	NC = cl(Pos,Neg,PT).
translateSNFClause(or(LL),NC) :-
	!,
	% start SNF clause
	convertLiterals(LL,now,FOLLiterals),
	splitLiterals(FOLLiterals,Pos,Neg),
	gensym(i,PT),
	NC = cl(Pos,Neg,PT).

%---------------------------------------------------------------------------
% deepModalTranslation(+MODALLITERALS,+FOLLITERALS,+VAR,-CLAUSES)
% Given a list of modal literals MODALLITERALS and a corresponding list
% of first-order literals FOLLITERALS, which we have obtained by a `shallow'
% translation of MODALLITERALS, where all elements of FOLLITERALS share the
% variable VAR, we return the `deep' translation of MODALLITERALS in CLAUSES.
% Basically, the `deep' translation of a modal literals is the translation 
% of the modal literal to first-order logic using de Nivelles translation 
% morphism (which includes adding clauses representing some of the axiom
% schemata valid for the modal operator of the modal literal under 
% consideration).

deepModalTranslation([],[],_,[]) :-
	!.
deepModalTranslation([box(M,A,L)|MLL],[FL|FLL],X,[NC|OtherClauses]) :-
	genAccPred(M,A,r,R),
	AccAtom =.. [R,X,Y],
	modTranslate(L,Y,FOLLiteral),
	gensym(m,PT),
	(FOLLiteral = not(L1) ->
	    NC = cl([],[FL,AccAtom,L1],PT)
	;
	    NC = cl([FOLLiteral],[FL,AccAtom],PT)
	),
	allAxiomClauses(box(M,A,L),FL,X,AxiomClauses),
	deepModalTranslation(MLL,FLL,X,TransClauses),
	append(AxiomClauses,TransClauses,OtherClauses).
deepModalTranslation([not(box(M,A,L))|MLL],[not(FL)|FLL],X,[NC1,NC2|OtherClauses]) :-
	genAccPred(M,A,r,R),
	genSkolemFunction(M,A,L,F),
	SkolemTerm =.. [F,X],
	AccAtom =.. [R,X,SkolemTerm],
	modTranslate(L,SkolemTerm,FOLLiteral),
	gensym(m,PT1),
	(FOLLiteral = not(_) ->
	    NC1 = cl([FL],[FOLLiteral],PT1)
	;
	    NC1 = cl([FL,FOLLiteral],[],PT1)
	),
	gensym(m,PT2),
	NC2 = cl([FL,AccAtom],[],PT2),
	allAxiomClauses(not(box(M,A,L)),not(FL),X,AxiomClauses),
	deepModalTranslation(MLL,FLL,X,TransClauses),
	append(AxiomClauses,TransClauses,OtherClauses).

%---------------------------------------------------------------------------
% genSkolemFunction(+MODALOPERATOR,+AGENT,+LITERAL,-FUNCTIONSYMBOL)
% generates a skolem function FUNCTIONSYMBOL which is uniquely associated
% with the triple (+MODALOPERATOR,+AGENT,+LITERAL).

genSkolemFunction(M,A,L,F) :-
	(skolemFunction(M,A,L,F) ->
	    !,
	    true
	;
	    gensym(f,F),
	    asserta(skolemFunction(M,A,L,F))
	),
	!.

%---------------------------------------------------------------------------
% modTranslate(+PROPLITERAL,+VAR,-FOLLITERAL)
% translates a propositional literal PROPLITERAL to a first-order literal
% FOLLITERAL with variable VAR.

modTranslate(not(P),X,not(Atom)) :-
	!,
	Atom =.. [P,X].	
modTranslate(P,X,Atom) :-
	atomic(P),
	!,
	Atom =.. [P,X].

%---------------------------------------------------------------------------
% generateModalAxioms(-CLAUSES)
% generates a list of clauses CLAUSES consisting of all the first-order
% correspondences for those modal axiom schemata which do not require a
% reference to the modal formulae under consideration under de Nivelle's
% translation morphism, that is, `d' and `t'.

generateModalAxioms(Clauses) :-
	findall(ml(M,A,Logic),modalLogic(M,A,Logic),Logics),
	generateModalAxioms(Logics,Clauses).

generateModalAxioms([],[]) :-
	!.
generateModalAxioms([ml(M,A,Logic)|LL],Clauses) :-
	(member(d,Logic) ->
	    genAccPred(M,A,r,R),
	    genAccPred(M,A,d,F),
	    SkolemTerm =.. [F,X],
	    AccAtom =.. [R,X,SkolemTerm],
	    gensym(ser,PT1),
	    Clause1 = [cl([AccAtom],[],PT1)]
	;
	    Clause1 = []
	),
	(member(t,Logic) ->
	    genAccPred(M,A,r,R),
	    AccAtom =.. [R,X,X],
	    gensym(refl,PT2),
	    Clause2 = [cl([AccAtom],[],PT2)]
	;
	    Clause2 = []
	),
	append(Clause1,Clause2,Clauses1),
	generateModalAxioms(LL,Clauses2),
	append(Clauses1,Clauses2,Clauses),
	!.

%---------------------------------------------------------------------------
% allAxiomClauses(+MODALLITERAL,+FOLLITERAL,+VAR,-CLAUSES)
% generates a list of clauses CLAUSES consisting of all the first-order
% correspondences for those modal axiom schemata which require a 
% reference to the modal formula MODALLITERAL under de Nivelle's
% translation morphism. FOLLITERAL is the first-order literal corresponding
% to MODALLITERAL under the `shallow' translation, where FOLLITERAL contains
% the variable VAR.

allAxiomClauses(box(M,A,_L),FL,X,ClauseList) :-
	(modalLogic(M,A,Logic) ->
	    axiomClausesLogic(Logic,box(M,A,_L),FL,X,ClauseList)
	;
	    ClauseList = []
	),
	!.
allAxiomClauses(not(box(_M,_A,_L)),_,_,[]) :-
	!.

axiomClausesLogic([],_,_,_,[]) :-
	!.
axiomClausesLogic([Ax|AxList],ML,FL,X,ClauseList) :-
	axiomClauseAxiom(Ax,ML,FL,X,ClauseList0),
	axiomClausesLogic(AxList,ML,FL,X,ClauseList1),
	append(ClauseList0,ClauseList1,ClauseList).
    
axiomClauseAxiom(b,box(M,A,L),FL1,Y,[Cl]) :-
	genAccPred(M,A,r,R),
	AccAtom =.. [R,X,Y],
	modTranslate(L,X,ML),
	gensym(sym,PT),
	(L = not(_) ->
	    Cl = cl([],[not(FL1),not(AccAtom),ML],PT)
	;
	    Cl = cl([ML],[not(FL1),not(AccAtom)],PT)
	),
	!.
axiomClauseAxiom(4,box(M,A,_L),FL1,X,[cl([FL2],[FL1,AccAtom],PT)]) :-
	genAccPred(M,A,r,R),
	modExtractPredicate(FL1,[P]),
	FL2 =.. [P,Y],
	AccAtom =.. [R,X,Y],
	gensym(trans,PT),
	!.
axiomClauseAxiom(5,box(M,A,_L),FL1,X,[cl([FL1],[FL2,AccAtom],PT)]) :-
	genAccPred(M,A,r,R),
	modExtractPredicate(FL1,[P]),
	FL2 =.. [P,Y],
	AccAtom =.. [R,X,Y],
	gensym(eucl,PT),
	!.
axiomClauseAxiom(_,_,_,_,[]) :-
	!.
	
%---------------------------------------------------------------------------
% shallowModalTranslation(+MODALLITERALS,+VAR,-FOLLITERALS)
% given a list of modal literals MODALLITERALS and a variable VAR we return
% the `shallow' translation of the literals in FOLLITERALS, where all the
% first-order literals in FOLLITERALS will have VAR as argument.
	
shallowModalTranslation([],_,[]) :- !.
shallowModalTranslation([box(M,A,L)|MLL],X,[FL|FLL]) :-
	genModalPred(M,A,L,MP),
	FL =.. [MP,X],
	shallowModalTranslation(MLL,X,FLL).
shallowModalTranslation([not(box(M,A,L))|MLL],X,[not(FL)|FLL]) :-
	genModalPred(M,A,L,MP),
	FL =.. [MP,X],
	shallowModalTranslation(MLL,X,FLL).

putBackOperator([],_,[]) :-
	!.
putBackOperator([L1|LL1],Op,[L2|LL2]) :-
	L2 =.. [Op,L1],
	putBackOperator(LL1,Op,LL2).

%---------------------------------------------------------------------------
% snfClassifyLiterals(+SNFLITERALS,-PROPLITERALS,OTHERLITERALS)
% given a list of literals SNFLITERALS, it return the list PROPLITERALS of 
% propositional literals occurring in SNFLITERALS, and a list OTHERLITERALS 
% of all other literals occurring in SNFLITERALS.

snfClassifyLiterals([],[],[]) :-
	!.
snfClassifyLiterals([sometime(L)|LL],PropositionalLiterals,[L|TemporalLiterals]) :-
	!,
	snfClassifyLiterals(LL,PropositionalLiterals,TemporalLiterals).
snfClassifyLiterals([next(L)|LL],PropositionalLiterals,[L|TemporalLiterals]) :-
	!,
	snfClassifyLiterals(LL,PropositionalLiterals,TemporalLiterals).
snfClassifyLiterals([box(M,A,L)|LL],PropositionalLiterals,[box(M,A,L)|ModalLiterals]) :-
	!,
	snfClassifyLiterals(LL,PropositionalLiterals,ModalLiterals).
snfClassifyLiterals([not(box(M,A,L))|LL],PropositionalLiterals,[not(box(M,A,L))|ModalLiterals]) :-
	!,
	snfClassifyLiterals(LL,PropositionalLiterals,ModalLiterals).
snfClassifyLiterals([not(L)|LL],[not(L)|PropositionalLiterals],TemporalLiterals) :-
	!,
	snfClassifyLiterals(LL,PropositionalLiterals,TemporalLiterals).
snfClassifyLiterals([L|LL],[L|PropositionalLiterals],TemporalLiterals) :-
	atomic(L),
	!,
	snfClassifyLiterals(LL,PropositionalLiterals,TemporalLiterals).

%---------------------------------------------------------------------------
% convertLiterals(+PROPLITERALS,+VAR,-FOLLITERALS)
% translates a list of propositional literals PROPLITERALS to a list of 
% first-order literals +FOLLITERALS, where all literals will share the 
% variable VAR.

convertLiterals([],_X,[]) :-
	!.
convertLiterals([L1|LL1],X,[L2|LL2]) :-
	convertLiteral(L1,X,L2),
	convertLiterals(LL1,X,LL2).

convertLiteral(not(P),X,not(A)) :-
	!,
	A =.. [P,X].
convertLiteral(P,X,A) :-
	atomic(P),
	A =.. [P,X].

%---------------------------------------------------------------------------
% getSometimeClauses(+FOLCLAUSES,-SOMETIMECLAUSES,+OTHERCLAUSES)
% given a list FOLCLAUSES of `pseudo' first-order clauses, returns the list
% SOMETIMECLAUSE of all sometime clauses in FOLCLAUSES and the list of
% OTHERCLAUSES of all other clauses in FOLCLAUSES.

getSometimeClauses([],[],[]) :-
	!.
getSometimeClauses([cl(PosLits,NegLits,PT)|CL],CLSometime,CLOther) :-
	!,
	getSometimeClauses(CL,CLSometime0,CLOther0),
	(member(sometime(_L),PosLits) ->
	    CLSometime = [cl(PosLits,NegLits,PT)|CLSometime0],
	    CLOther    = CLOther0
	;
	    CLSometime = CLSometime0,
	    CLOther    = [cl(PosLits,NegLits,PT)|CLOther0]
	).
getSometimeClauses([cl(PosLits,NegLits)|CL],CLSometime,CLOther) :-
	!,
	getSometimeClauses(CL,CLSometime0,CLOther0),
	(member(sometime(_L),PosLits) ->
	    CLSometime = [cl(PosLits,NegLits,PT)|CLSometime0],
	    CLOther    = CLOther0
	;
	    CLSometime = CLSometime0,
	    CLOther    = [cl(PosLits,NegLits,PT)|CLOther0]
	).
	    
%---------------------------------------------------------------------------
% mainProofProcedure(+CLAUSEINDEX1,+USABLES,+SOMETIMECLAUSES,-CONSISTENT,
%                    -SATURATED,-STAT)
% the main proof procedure: takes as input an (empty) clause index
% CLAUSEINDEX1, a list USABLES of usable `pseudo' first-order clauses,
% a list SOMETIMESCLAUSES of sometime clauses. Returns CONSISTENT which
% will be `true' if (USABLES & SOMETIMECLAUSES) is consistent and `false'
% otherwise, a list SATURATED of clauses, and some statistical information
% in STAT.
% Performs forward subsumption on USABLES, orders the remaining clauses 
% according to the number of symbols they contain, and calls the actual
% resolution prover.

mainProofProcedure([ST0,SL0],CL1,CLSometime,Consistent,CL,statistics(NF,NR,NS,NTRSuccess,NTRFail,NTRNR,NC)) :-
%       delete redundant clauses from CL1
	modForwardSubsumption(CL1,[ST0,SL0],[],stat(0,0),
	                      [ST1,SL1],CL2,stat(NR1,NS1)),
%       compute resolvents beween clauses in CL2 and clauses in NST
	ifOption(debugLevel,DL,(DL>7),(
	         write('Forward subsumption finished.'), nl,
		 write('Productive new clauses are: '), nl,
                 printClauses(spass,debug,CL2),
		 write('Starting new level of computation.'),nl)),
        keycreate(CL2,SCL),
	keysort(SCL,SCL2),
	resolutionProver([ST1,SL1],SCL2,CLSometime,stat(0,NR1,NS1,0,0,0),
                         [_ST2,_SL2],CL,Consistent,stat(NF,NR,NS,NTRSuccess,NTRFail,NTRNR)),
	length(CL,NC).

%---------------------------------------------------------------------------
% resolutionProver(+CLAUSEINDEX1,+USABLES,+SOMETIMECLAUSES,+STAT1,
%                  -CLAUSEINDEX2,-SATURATED,-CONSISTENT,-STAT2)
% Takes as input a clause index CLAUSEINDEX1, a list USABLES of usable 
% `pseudo' first-order clauses, a list SOMETIMESCLAUSES of sometime clauses,
% and some statistical information about the derivation up to this point.
% Returns a modified clause index CLAUSEINDEX2, a list SATURATED of clauses, 
% CONSISTENT which will be `true' if (CLAUSEINDEX1 & USABLES & SOMETIMECLAUSES) 
% is consistent and `false' otherwise, and updated statistical information.
% resolutionProver uses inferenceLoop to perform inference steps corresponding
% to initial resolution, step resolution, propositional resolution, and modal
% resolution steps. If at this point we already derive a contradiction, we 
% are done. Otherwise, we will try to apply the temporal resolution to one of
% the sometime clauses using tryTRForEachSometimeClause.

resolutionProver([ST1,SL1],UsableClauses,CLSometime,stat(NF1,NR1,NS1,NTRS1,NTRF1,NTRNR1),
                 [ST2,SL2],FinalClauses,Consistent,stat(NF,NR,NS,NTRS,NTRF,NTRNR)) :-
	inferenceLoop([ST1,SL1],UsableClauses,stat(NF1,NR1,NS1),
                        [ST2,SL2],SaturatedClauses,Consistent0,stat(NF2,NR2,NS2)),
	(Consistent0 == false ->
	    !,
	    NF = NF2,
	    NR = NR2,
	    NS = NS2,
	    NTRS  = NTRS1,
	    NTRF  = NTRF1,
	    NTRNR = NTRNR1,
	    FinalClauses = SaturatedClauses,
	    Consistent = Consistent0
	;
	    ifOption(debugLevel,DL,(DL>2),(
	         write('Clauses saturated. Will try to apply temporal resolution next.'), nl,
		 write('Current clauses: '), nl,
                 printClauses(spass,debug,SaturatedClauses),
		 write('Starting application of temporal resolution.'),nl)),
	    !,
	    tryTRForEachSometimeClause(CLSometime,CLSometime,[ST2,SL2],stat(NF2,NR2,NS2,NTRS1,NTRF1,NTRNR1),SaturatedClauses,
	                    FinalClauses,Consistent,stat(NF,NR,NS,NTRS,NTRF,NTRNR))
	),
	!.

%---------------------------------------------------------------------------
% tryTRForEachSometimeClause(+USABLESOMETIMECLAUSES,+SOMETIMECLAUSES,
%    +CLAUSEINDEX1,+STAT1,+SATURATED1,-SATURATED2,-CONSISTENT,-STAT2)
% USABLESOMETIMECLAUSES is the list of sometimes clauses to which we not yet
% tried to apply the temporal resolution rule. SOMETIMECLAUSES is the list of
% all sometime clauses. CLAUSEINDEX1, STAT1, SATURATED1 are a clause index, 
% statistical information, and a set of clauses, respectively. We return a
% saturated clause set SATURATED2, CONSISTENT, and updated statistical 
% information. 
% If USABLESOMETIMECLAUSES is empty, then we were unsuccessful to apply the
% temporal resolution rule to any of the sometime clauses. Thus, SATURATED1
% is a consistent saturated set of clauses. Otherwise, we try to apply to
% the first sometime clause in USABLESOMETIMECLAUSES using tryTR. If the
% set of temporal resolvents tryTR is empty, then we call tryTRForEachSometimeClause
% for the remaining clauses in USABLESOMETIMCLAUSES. Otherwise, we call
% resolutionProver with the temporal resolvents as the new set of 
% usable clauses.

tryTRForEachSometimeClause([],_,[_ST1,_SL1],stat(NF2,NR2,NS2,NTRS2,NTRF2,NTRNR2),SaturatedClauses,
	        SaturatedClauses,true,stat(NF2,NR2,NS2,NTRS2,NTRF2,NTRNR2)) :-
	!.
tryTRForEachSometimeClause([SCl1|SClList],CLSometime,[ST1,SL1],stat(NF2,NR2,NS2,NTRS2,NTRF2,NTRNR2),SaturatedClauses,
                FinalClauses,Consistent,stat(NF,NR,NS,NTRS,NTRF,NTRNR)) :-
        garbage_collect,
	!,
	tryTR(SCl1,[ST1,SL1],stat(NF2,NR2,NS2,NTRS2,NTRF2,NTRNR2),
                         Resolvents,stat(NF3,NR3,NS3,NTRS3,NTRF3,NTRNR3)),
	!,
	(Resolvents == [] ->
	    !,
	    tryTRForEachSometimeClause(SClList,CLSometime,[ST1,SL1],stat(NF3,NR3,NS3,NTRS3,NTRF3,NTRNR3),SaturatedClauses,
                            FinalClauses,Consistent,stat(NF,NR,NS,NTRS,NTRF,NTRNR))
	;
	    !,
	    resolutionProver([ST1,SL1],Resolvents,CLSometime,stat(NF3,NR3,NS3,NTRS3,NTRF3,NTRNR3),
                             [_ST2,_SL2],FinalClauses,Consistent,stat(NF,NR,NS,NTRS,NTRF,NTRNR))
	).
	
%---------------------------------------------------------------------------
% tryTR(+SOMETIMECLAUSE,+CLAUSEINDEX1,+STAT1,-LOOPCLAUSES,-STAT2)
% SOMETIMECLAUSE is a sometime clause, CLAUSEINDEX a clause index, STAT1 
% statistical information about the derivation up to now. We try to apply
% the temporal resolution rule to the SOMETIMECLAUSE. If successful, we
% will return a non-empty set of resolvents LOOPCLAUSES. Otherwise 
% LOOPCLAUSES will be the empty set. In both cases we return updated 
% statistical information.
% The actual derivation of temporal resolvents is performed by searchForLoop.
% The clauses returned by searchForLoop will be subjected to backward
% subsumption which ensures that we only consider an application of the temporal
% resolution rule as successful if it returns some new clauses.

tryTR(SCl1,[ST1,SL1],stat(NF1,NR1,NS1,NTRS1,NTRF1,NTRNR1),
                 LoopClauses,stat(NF3,NR3,NS3,NTRS3,NTRF3,NTRNR3)) :-
%       there seem to be no side-effects on the clause index, so copying it
%       doesn't seem to be necessary. uh 09.10.2001
%	copy_term([ST1,SL1],[ST0,SL0]),
        ST0 = ST1,
	SL0 = SL1,
	(SCl1 = cl([sometime(L)],Neg,_PT) ->
	    NegativeLiterals = Neg,
	    term_variables(L,[X]),
	    % do I have to lift this literal? uh 07.06.2001
% uh 24.10.2008: It seems that using a variable X leads to incorrect
%                to be found. So, we now use the constant `now' instead
%	    liftLiterals([L],X,[SometimeLiteral])
	    liftLiterals([L],loop(X),[SometimeLiteral])
	;
	    write('WARNING: Sometime clause '), nl,
	    printClauses(spass,debug,SCl1),
	    write('has not the required form'),
	    !,
	    fail
	),
	ifOption(debugLevel,DL,(DL>2),(write('Loop search in '),write(SometimeLiteral),nl)),
	searchForLoop([ST0,SL0],[],SometimeLiteral,stat(NF1,NR1,NS1,NTRS1,NTRF1),
                  ResultingClauses,stat(NF3,NR2,NS2,NTRS3,NTRF3)),
	ifOption(debugLevel,DL,(DL>2),(write('Loop search in '),write(SometimeLiteral),nl,
	         printLoopClauses(ResultingClauses))),
	generateLoopClauses(ResultingClauses,SometimeLiteral,NegativeLiterals,
                            DerivedLoopClauses),
	modBackwardSubsumption(DerivedLoopClauses,SL0,stat(NR2,NS2),
                               NonSubsumedLoopClauses,stat(NR3,NS3)),
	(NonSubsumedLoopClauses == [] ->
	    NTRNR3 = NTRNR1
	;
	    NTRNR3 is NTRNR1+1
	),
	ifOption(debugLevel,DL,(DL>2),(write('Loop search in '),write(SometimeLiteral),
                 write(' produced clauses as resolvents by temporal resolution'),nl,
                 printClauses(spass,debug,NonSubsumedLoopClauses),nl,
                 write(NTRS1), write(NTRF1), write(NTRS3), write(NTRF3),nl)),
        keycreate(NonSubsumedLoopClauses,LoopClausesWithKeys),
	keysort(LoopClausesWithKeys,LoopClauses),
	!.

%---------------------------------------------------------------------------
% additionalClausesBFS(+CLAUSES,+LITERAL,-CONJECTURES)
% Generates the conjecture clauses we like to prove in a particular iteration
% of loop search. If CLAUSES is a non-empty list of clauses, then we have to
% add LITERAL to each of the clauses to get the list of conjecture clauses.
% (The construction of these clauses is performed by additionalClausesBFS0;
% the difficulty is to ensure that LITERAL and the rest of the clause share
% the same variable.) Otherwise the unit clause consisting of LITERAL is the 
% only conjecture clause.  

additionalClausesBFS([],SometimeLiteral,[Clause1]) :-
	computeCharacteristic(cl([SometimeLiteral]),Clause1),
	!.
additionalClausesBFS([C1|CL],SometimeLiteral,Clauses) :-
	additionalClausesBFS0([C1|CL],SometimeLiteral,Clauses),
	!.

additionalClausesBFS0([],_,[]) :-
	!.
additionalClausesBFS0([_-cl(Literals,LL)|CL],SometimeLiteral,[Clause1|CL1]) :-
	!,
	additionalClausesBFS0([cl(Literals,LL)|CL],SometimeLiteral,[Clause1|CL1]).
additionalClausesBFS0([cl(Literals,_)|CL],SometimeLiteral,[Clause1|CL1]) :-
	term_variables(SometimeLiteral,[X]),
	term_variables(Literals,[X]),
	% do I have to lift these literals? uh 07.06.2001
% uh 24.10.2008: It seems that using a variable X leads to incorrect
%                to be found. So, we now use the constant `now' instead
%	liftLiterals(Literals,X,LiftedLiterals),
	liftLiterals(Literals,loop(X),LiftedLiterals),
	computeCharacteristic(cl([SometimeLiteral|LiftedLiterals]),Clause1),
	additionalClausesBFS0(CL,SometimeLiteral,CL1).


%---------------------------------------------------------------------------
% mutuallySubsumingSets(+CLAUSES1,+CLAUSES2)
% succeeds if the two sets of clauses CLAUSES1 and CLAUSES2 subsume each
% other.

mutuallySubsumingSets(Set1,Set2) :-
	ifOption(debugLevel,DL,(DL>7),(write('Checking whether two clause sets are mutually subsuming'), nl)),
	modBackwardSubsumption(Set1,Set2,stat(0,0),[],stat(_,_)),
	modBackwardSubsumption(Set2,Set1,stat(0,0),[],stat(_,_)),
	ifOption(debugLevel,DL,(DL>7),(write('successful'), nl)),
	!.
	
%---------------------------------------------------------------------------
% searchForLoop(+CLAUSEINDEX1,+LOOPCLAUSES1,+SOMETIMELITERAL,+STAT1,
%               -LOOPCLAUSES2,-STAT2)
% Performs the actual breath-first search for a set of loop clauses for
% the eventuality SOMETIMELITERAL. LOOPCLAUSE1 is a list of clauses representing
% our current approximation of a potential loop, while LOOPCLAUSES2 is the
% final list of loop clauses or the value `true' if SOMETIMELITERAL cannot
% be true, or the value `false' if there is no loop. See the full version of
% the FroCoS paper for a description.

searchForLoop([ST0,SL0],OldClauses,SometimeLiteral,stat(NF1,NR1,NS1,NTRS1,NTRF1),
              Result,stat(NF4,NR4,NS4,NTRS4,NTRF4)) :-
        ifOption(debugLevel,DL,(DL>2),(write('Loop search current node '), nl, 
                                       printClauses(spass,debug,OldClauses),nl)),
	additionalClausesBFS(OldClauses,SometimeLiteral,STLClauses),
        ifOption(debugLevel,DL,(DL>2),(write('Loop search clauses '), nl, 
                                       printClauses(spass,debug,STLClauses),nl)),
	keycreate(STLClauses,STLClauses1),
	keysort(STLClauses1,STLClauses2),
	inferenceBFS(STLClauses2,[ST0,SL0],[],stat(NF1,NR1,NS1),
                     [_ST2,_SL2],FinalClauses,stat(NF2,NR2,NS2),Bool),
	(Bool == false ->
	    % loop in true
	    Result = true,
	    NF4 = NF2, NR4 = NR2, NS4 = NS2,
            NTRS4 is NTRS1+1,
            NTRF4  = NTRF1
	;
	    (FinalClauses == [] ->
		% no loop found
		Result = false,
		NF4 = NF2, NR4 = NR2, NS4 = NS2,
                NTRS4 = NTRS1,
                NTRF4 is NTRF1+1
	    ;
		(mutuallySubsumingSets(FinalClauses,OldClauses) ->
		    % loop found
		    ifOption(debugLevel,DL,(DL>2),(write('loop found'), nl)),
		    Result = OldClauses,
		    NF4 = NF2, NR4 = NR2, NS4 = NS2,
		    NTRS4 is NTRS1+1,
		    NTRF4  = NTRF1
		;
		    ifOption(debugLevel,DL,(DL>2),(write('no loop yet'), nl)),
		    !,
		    searchForLoop([ST0,SL0],FinalClauses,SometimeLiteral,stat(NF2,NR2,NS2,NTRS1,NTRF1),
                                  Result,stat(NF4,NR4,NS4,NTRS4,NTRF4))
		)
	    )
	).

%---------------------------------------------------------------------------
% modEmptyClause(+CLAUSES)
% succeeds if the list of clauses CLAUSES contains the empty clause
	    
modEmptyClause(UsableClauses) :-
	(member(cl([],_PT),UsableClauses) ; member(_N-cl([],_PT),UsableClauses)),
	 !.

%---------------------------------------------------------------------------
% flatClause(+CLAUSE,+VAR)
% succeeds if CLAUSE is a clause of the form (-)p1(VAR) | ... | (-)pn(VAR)

flatClause(_N-Clause,X) :-
	flatClause(Clause,X).
flatClause(cl([],_Char),_X) :-
	!.
flatClause(cl([L|LL],Char),X) :-
	(L = not(A) ->
	    Atom = A
	;
	    Atom = L
	),	
	arg(1,Atom,Z),
	(Z == X ; Z == loop(X)),
	!,
	flatClause(cl(LL,Char),X).	    

%---------------------------------------------------------------------------
% classifyDerivedClauses(+CLAUSES,-DEEPCLAUSES,-FLATCLAUSES)
% splits a list CLAUSES of clauses into a list FLATCLAUSES of the form
%  (-)p1(VAR) | ... | (-)pn(VAR) and a list DEEPCLAUSES of all other
% clauses.

classifyDerivedClauses([],[],[],[]) :-
	!.
classifyDerivedClauses([C|CL],DeepClauses,FlatClauses,GroundClauses) :-
	classifyDerivedClauses(CL,DeepClauses1,FlatClauses1,GroundClauses1),
	(ground(C) ->
	    DeepClauses = DeepClauses1,
	    FlatClauses = FlatClauses1,
	    GroundClauses = [C|GroundClauses1]
	;
	    term_variables(C,[X]),
	    (flatClause(C,X) ->
		DeepClauses = DeepClauses1,
		FlatClauses = [C|FlatClauses1],
		GroundClauses = GroundClauses1
	    ;
		DeepClauses = [C|DeepClauses1],
		FlatClauses = FlatClauses1,
		GroundClauses = GroundClauses1
	    )
	).

filterGroundClauses([],[],[]) :-
	!.
filterGroundClauses([C|CL],NonGroundClauses,GroundClauses) :-
	filterGroundClauses(CL,NonGroundClauses1,GroundClauses1),
	(ground(C) ->
	    NonGroundClauses = NonGroundClauses1,
	    GroundClauses = [C|GroundClauses1]
	;
	    NonGroundClauses = [C|NonGroundClauses1],
	    GroundClauses = GroundClauses1
	).


%---------------------------------------------------------------------------
% inferenceBFS(+USABLES,+CLAUSEINDEX1,+NEWCLAUSES,+STAT1,
%              -CLAUSEINDEX2,-FINALCLAUSES,-STAT2,-CONSISTENT)
% Derives the next approximation of the set of loop clauses. The problem is
% that we only want to perform inference steps corresponding to applications
% of step resolution. However, if we saturate a given list of clauses under
% ordered resolution, then we will also perform inference steps corresponding
% to initial resolution and propositional resolution. To avoid this, after 
% application of ordered resolution to a particular given clause, we split
% the derived clauses into `flat' clauses (that is, propositional clauses)
% and deep clauses (that is, step clauses). Only to the deep clauses we will
% continue to apply ordered resolution to get rid of all the deep literals.

inferenceBFS([],[ST0,SL0],NewClauses,stat(NF1,NR1,NS1),
	     [ST0,SL0],NewClauses,stat(NF1,NR1,NS1),true) :-
        !.
inferenceBFS([Given|UsableClauses],[ST0,SL0],NewClauses,stat(NF1,NR1,NS1),
	     [ST3,SL3],FinalClauses,stat(NF5,NR5,NS5),Consistent) :-
        !,
%       computing all factors and resolvents between Given and clauses in ST0
	modConsequences([ST0,SL0],[Given],stat(NF1,NR1),tres,
                        [ST1,SL1],Consequences,stat(NF2,NR2)),
	ifOption(debugLevel,DL,(DL>7),(write('Starting subsumption tests in inferenceBFS for'), 
                                       nl, printClauses(spass,debug,Consequences))),
%       delete clauses from Consequences which are subsumed by clauses in ST,
%       the result is NonSubsumedConsequences1
	append(SL1,NewClauses,OldClauses),
	!,
	simpleReduction(Consequences,OldClauses,[],stat(NR2,NS1),
	                NonSubsumedConsequences1,stat(NR3,NS3)),
	simpleReduction(NewClauses,NonSubsumedConsequences1,[],stat(NR3,NS3),
	                NonSubsumedNewClauses,stat(NR4,NS4)),
	!,
%       Enter next level of the saturation process if necesary
	ifOption(debugLevel,DL,(DL>2),(write('Derived clauses so far: '), nl,
            	 printClauses(spass,debug,NonSubsumedConsequences1))),
	(modEmptyClause(NonSubsumedConsequences1) ->
	    Consistent   = false,
	    FinalClauses = [],
	    NF5 = NF2, NR5 = NR4, NS5 = NS4
	;
	    classifyDerivedClauses(NonSubsumedConsequences1,
                                   _DeepClauses,FlatClauses,_),
	    append(NonSubsumedNewClauses,FlatClauses,NewClauses2),
%           old version uh 04.10.2001
%	    keycreate(DeepClauses,DeepClauses2),
	    filterGroundClauses(NonSubsumedConsequences1,NonGroundClauses,_),
	    keycreate(NonGroundClauses,DeepClauses2),
	    keysort(DeepClauses2,DeepClauses3),
	    keymerge(UsableClauses,DeepClauses3,UsableClauses2),
	    !,
	    inferenceBFS(UsableClauses2,[ST1,SL1],NewClauses2,stat(NF2,NR4,NS4),
 		         [ST3,SL3],FinalClauses,stat(NF5,NR5,NS5),Consistent)
	),
	!.
%inferenceBFS([Given|UsableClauses],[ST0,SL0],NewClauses,stat(NF1,NR1,NS1),
%	     [ST3,SL3],FinalClauses,stat(NF5,NR5,NS5),Consistent) :-
%        inferenceBFS([0-Given|UsableClauses],[ST0,SL0],NewClauses,stat(NF1,NR1,NS1),
%	             [ST3,SL3],FinalClauses,stat(NF5,NR5,NS5),Consistent).

%---------------------------------------------------------------------------
% generateLoopClauses(+RESULT,+SOMETIMELITERAL,+LITERALS,-LOOPCLAUSES)
% generates the actual clauses resulting from an (potentially failed) 
% application of temporal resolution to the clause 
%            LITERALS | sometime(SOMETIMELITERAL)
% RESULT is the result from inferenceBFS. If RESULT is `false' then then we 
% haven't found a loop and LOOPCLAUSES is the empty list of clauses. If RESULT
% is `true', the `true' represents the loop and we construct the resolvents
% accordingly. Otherwise, RESULT is a list of clauses representing the loop
% we have found and we construct the resolvents from them.

generateLoopClauses(true,SometimeLiteral,NegLiterals,LoopClauses) :-
	modNegateLiterals(NegLiterals,NegatedLiterals),
	modExtractPredicate(SometimeLiteral,[P]),
	NextP =.. [P,f(X)],
	NowP  =.. [P,X],
	(SometimeLiteral = not(_) ->
	    NextL = not(NextP),
	    NowL  = not(NowP)
	; 
	    NextL = NextP,
	    NowL  = NowP
	),
	Clause1 = cl([NowL|NegatedLiterals]),
	genWaitingPred(SometimeLiteral,PW),
	term_variables(SometimeLiteral,[X]),
	Waiting =.. [PW,X],
	Clause2 = cl([not(Waiting),NextL]),
	computeCharacteristic([Clause1,Clause2],LoopClauses).
generateLoopClauses(false,_SometimeLiteral,_NegLiterals,[]) :-
	!.
generateLoopClauses([],_SometimeLiteral,_NegLiterals,[]) :-
	!.
generateLoopClauses([cl(Literals,_)|CL],SometimeLiteral,NegLiterals,LoopClauses) :-
	generateLoopClauses(CL,SometimeLiteral,NegLiterals,LoopClauses0),
	modNegateLiterals(NegLiterals,NegatedLiterals),
	term_variables(SometimeLiteral,[X]),
	term_variables(Literals,[X]),
	modExtractPredicate(SometimeLiteral,[P]),
	NextP =.. [P,f(X)],
	NowP  =.. [P,X],
	(SometimeLiteral = not(_) ->
	    NextL = not(NextP),
	    NowL  = not(NowP)
	; 
	    NextL = NextP,
	    NowL  = NowP
	),
	append([NowL|NegatedLiterals],Literals,Literals1),
	Clause1 = cl(Literals1),
	genWaitingPred(SometimeLiteral,PW),
	Waiting =.. [PW,X],
	liftLiterals(Literals,X,LiftedLiterals),
	Clause2 = cl([not(Waiting),NextL|LiftedLiterals]),
	computeCharacteristic([Clause1,Clause2],LoopClauses1),
	append(LoopClauses1,LoopClauses0,LoopClauses),
	!.
generateLoopClauses([_N-cl(Literals,_)|CL],SometimeLiteral,NegLiterals,LoopClauses) :-
	generateLoopClauses([cl(Literals,_)|CL],SometimeLiteral,NegLiterals,LoopClauses),
	!.

%---------------------------------------------------------------------------
% liftLiterals(+FLATLITERALS,+VAR,-DEEPLITERALS)
% given a list of literals of the form (-)p(X) returns a list of literals
% of the form (-)p(f(VAR)).

liftLiterals([],_X,[]) :-
	!.
liftLiterals([not(L0)|LL0],X,[not(L1)|LL1]) :-
	liftLiterals(LL0,X,LL1),
	functor(L0,P,_),
	L1 =.. [P,f(X)].
liftLiterals([L0|LL0],X,[L1|LL1]) :-
	liftLiterals(LL0,X,LL1),
	functor(L0,P,_),
	L1 =.. [P,f(X)].

%---------------------------------------------------------------------------
% inferenceLoop(+CLAUSEINDEX1,+USABLES,+STAT1,-CLAUSEINDEX2,-FINALCLAUSES,
%               -CONSISTENT,-STAT2)
% implements the main inference loop of a resolution theorem prover.
% Given a clause index CLAUSEINDEX1, a list USABLES of clauses, and some
% statistical information STAT1 about the derivation so far, we proceed
% as follows. If USABLES contains the empty clause, then we are done. 
% Otherwise we pick the first clause in USABLES, compute all the resolvents
% between this given clause and the clauses in CLAUSEINDEX1, perform
% redundancy elimination, add the remaining derived clauses to USABLES
% and call inferenceLoopCont (which in turn will call inferenceLoop with
% an updated list of USABLES.

inferenceLoop([ST,SL],UsableClauses,Stat,[ST,SL],FinalClauses,Bool,Stat) :-
	member(cl([],_Pt),UsableClauses),
	!,
	Bool = false,
	append(SL,UsableClauses,FinalClauses).
inferenceLoop([ST,SL],UsableClauses,Stat,[ST,SL],FinalClauses,Bool,Stat) :-
	member(_N-cl([],_Pt),UsableClauses),
	!,
	Bool = false,
	append(SL,UsableClauses,FinalClauses).
inferenceLoop([ST0,SL0],[_N-Given|UsableClauses],stat(NF1,NR1,NS1),
	        [ST3,SL3],FinalClauses,Consistent,stat(NF5,NR5,NS5)) :-
        !,
%       computing all factors and resolvents between Given and clauses in ST0
	modConsequences([ST0,SL0],[Given],stat(NF1,NR1),sres,
                        [ST1,SL1],Consequences,stat(NF2,NR2)),
	ifOption(debugLevel,DL,(DL>7),(write('Starting subsumption tests in inferenceLoop for'), nl, printClauses(spass,debug,Consequences))),
%       delete clauses from Consequences which are subsumed by clauses in ST,
%       the result is NonSubsumedConsequences1
	modForwardSubsumption(Consequences,[ST1,SL1],[],stat(NR2,NS1),
	                      [ST2,SL2],NonSubsumedConsequences1,stat(NR3,NS3)),
%       perform forward and backward subsumption between clauses in UsableClauses 
%       and NonSubsumedConsequences1, the result is 
        modForwardList(NonSubsumedConsequences1,UsableClauses,[],stat(NR3,NS3),
                       NonSubsumedUsableClauses,
	               NonSubsumedConsequences2,stat(NR4,NS4)),
%       Enter next level of the saturation process if necesary
	!,
	inferenceLoopCont(NonSubsumedConsequences2,NonSubsumedUsableClauses,
                         [ST2,SL2],stat(NF2,NR4,NS4),
			 [ST3,SL3],FinalClauses,Consistent,stat(NF5,NR5,NS5)),
	!.

%---------------------------------------------------------------------------
% inferenceLoopCont(+NON_SUBSUMED_CONSEQUENCES,+US_CLAUSES,
%                  +OLD_WO_INDEX,+OLD_STAT,
%                  -NEW_WO_INDEX,-FINAL_CLAUSESET,-CONSISTENT,-NEW_STAT)
% Decides whether a new level of saturation has to be computed:
% If NON_SUBSUMED is empty, the saturation process is finished
% and the clause set is consistent (therefore, CONSISTENT will be true).
% Otherwise we go back to inferenceLoop and compute the next saturation
% level.

inferenceLoopCont([],[],[ST,SL],STAT,
                       [ST,SL],SL,true,STAT) :-
	ifOption(clauseOutput,yes,(write('Final Clauses: '),nl,printClauses(list,_,SL))),
	!.
inferenceLoopCont([],[C|CL],[ST,SL],stat(NF1,NR1,NS1),
                 [ST2,SL2],FinalClauses,Consistent,STAT2) :-
	!,
	ifOption(debugLevel,DL,(DL>3),(
	         write('No new clauses. Remaining clauses are:'), nl,
                 printClauses(spass,debug,[C|CL]),
		 write('Starting new level of computation.'),nl)),
	inferenceLoop([ST,SL],[C|CL],stat(NF1,NR1,NS1),
                        [ST2,SL2],FinalClauses,Consistent,STAT2).
inferenceLoopCont([C|CL1],CL2,[ST,SL],STAT1,
                 [ST2,SL2],FinalClauses,Consistent,STAT2) :-
        !,
	keymerge([C|CL1],CL2,CL3),
	ifOption(debugLevel,DL,(DL>3),(
	         write('After merging clauses:'), nl,
                 printClauses(spass,debug,CL3), nl,
		 write('Starting new level of computation.'),nl)),
	inferenceLoop([ST,SL],CL3,STAT1,[ST2,SL2],FinalClauses,Consistent,STAT2),
	!.

keycreate([],[]) :-
	!.
keycreate([cl(Literals,char(PT,VL,PL,FL,W))|L],
	       [W-cl(Literals,char(PT,VL,PL,FL,W))|WL]) :-
        keycreate(L,WL),
	!.
keycreate([W-CL|L],[W-CL|WL]) :-
        keycreate(L,WL),
	!.

keymerge([],L1,L2) :-
	keycreate(L1,L2),
	!.
keymerge(L1,[],L2) :-
	keycreate(L1,L2).
keymerge([M-A|L1],[N-B|L2],[E|L3]) :-
	(M < N ->
	    E = M-A,
	    keymerge(L1,[N-B|L2],L3)
	;
	    E = N-B,
	    keymerge([M-A|L1],L2,L3)
	),
	!.
keymerge([cl(LL,char(PT,VL,PL,FL,M))|L1],[N-B|L2],L3) :-
	keymerge([M-cl(LL,char(PT,VL,PL,FL,M))|L1],[N-B|L2],L3),
	!.
keymerge([M-A|L1],[cl(LL,char(PT,VL,PL,FL,N))|L2],L3) :-
	keymerge([M-A|L1],[N-cl(LL,char(PT,VL,PL,FL,N))|L2],L3),
	!.
keymerge([cl(LL1,char(PT1,VL1,PL1,FL1,M))|L1],
	 [cl(LL2,char(PT2,VL2,PL2,FL2,N))|L2],L3) :-
	keymerge([M-cl(LL1,char(PT1,VL1,PL1,FL1,M))|L1],
	         [N-cl(LL2,char(PT2,VL2,PL2,FL2,N))|L2],L3),
	!.

spassLoop([[OSTP,OSTN],OSL],[],_,stat(NF1,NR1),
          [[OSTP,OSTN],OSL],[],stat(NF1,NR1)) :-
        !.
spassLoop([[STP1,STN1],SL1],[C1|CL1],Sorted,stat(NF1,NR1),
          [[STP3,STN3],SL3],Consequences,stat(NF3,NR4)) :-
        (Sorted == false ->
	    keycreate([C1|CL1],WeightedClauses),
	    keysort(WeightedClauses,[Given|SortedClauses])
	;
	    Given = C1,
	    SortedClauses = CL1
	),
	modConsequences([[STP1,STN1],SL1],[Given],stat(NF1,NR1),sres,
	                [[STP2,STN2],SL2],Consequences,stat(NF2,NR2)),
	simpleReduction(Consequences,_OldClauses,[],stat(NR2,NR2),
                        NonSubsumedConsequences,stat(NR3,NR3)),
        keycreate(NonSubsumedConsequences,WeightedConsequences),
	keysort(WeightedConsequences,SortedConsequences),
	keymerge(SortedClauses,SortedConsequences,SortedUsables),
	!,
	spassLoop([[STP2,STN2],SL2],SortedUsables,stat(NF2,NR3),
                  [[STP3,STN3],SL3],Consequences,stat(NF3,NR4)).

%---------------------------------------------------------------------------
% modConsequences(+OLD_WO_INDEX,+US_CLAUSES,+OLD_STATISTICS,+RES_TYPE,
%                 -NEW_WO_INDEX,-DERIVED_CLAUSES,-NEW_STATISTICS)
% Computes resolvents between clauses in the union NEWCLAUSES and factors of 
% NEWCLAUSES and clauses in OLDCLAUSES.

modConsequences([[OSTP,OSTN],OSL],Clauses,stat(NF1,NR1),ResType,
	        [[NSTP,NSTN],NSL],Consequences,stat(NF2,NR2)) :-
%       compute the clause structures for NewClauses
	all_maxlits_all_clauses(Clauses,[],CLMax1),
	(CLMax1 == [] ->
	    write('WARNING '),nl,
	    printClauses(spass,debug,Clauses),
	    write('produced no maximal literals'),nl,
	    all_maxlits_all_clauses(Clauses,[],CLMax1)
	;
	    true
	),
%       compute all factors of clauses in NewClauses
 	modSetfactor(CLMax1,[],Factors),       
%       remove tautological factors
        modRemoveTautologies(Factors,NonTautologicalFactors),
	length(NonTautologicalFactors,NF),
	NF2 is NF1+NF,
	append(CLMax1,NonTautologicalFactors,CLMax2),
%       put the clause structures into the substitution tree
	st_insertClauses(CLMax2,[OSTP,OSTN],[NSTP,NSTN]),
%       put the clauses corresponding to the factors into the 
%       set of all clauses
	structuresToClauses(NonTautologicalFactors,[],FactorSet),
	append(Clauses,FactorSet,CFL),
	append(CFL,OSL,NSL),
%       compute the resolvents (and removes the tautologies pmb)
%       Important: CLMax2 contains the new clauses accessible via positive and
%                         negative maximal literals.
%                  OSTP   contains the old clauses accessible via positive 
%                         maximal literals.
%                  NSTP   contains the old and new clauses accessible via 
%                         positive maximal literals.
%                  NSTN   contains the old and new clauses accessible via 
%                         negative maximal literals.
%       In the following call to modSetproduct OSTP and NSTN are used. 
%       The intention is that all clauses in CLMax2 are resolved with clauses 
%       in OSTP and NSTN. Using NSTP instead of OSTP would result in duplicate
%        computation of resolvents.
	ifOption(debugLevel,DL,(DL>3),(write('Starting modSetproduct with '), write(CLMax2),nl)),
	modSetproduct(CLMax2,[OSTP,NSTN],[],ResType,Consequences),
	length(Consequences,NR),
	NR2 is NR1+NR,
	ifOption(debugLevel,DL,(DL>3),(write('Finished modSetproduct: '),nl,
	                               printClauses(spass,debug,Consequences))),
	!.


%---------------------------------------------------------------------------
% modSetproduct(+CLAUSELIST1,+CLAUSE,+MAXLITS,-RESOLVENTS)
% RESOLVENTS is the list of all resolvents of clauses in CLAUSELIST1 with 
% CLAUSE (which contains MAXLITS as maximal literals) with respect to 
% order-restricted resolution.

modSetproduct([], _ST, Acc, _,Acc) :-
	!.
modSetproduct([not(L)-Clause|Tail],[OSTEP,NSTN],Acc1,ResType,ResolventSet) :-
	findall(Resolvent,maximalResolvent(L,neg,Clause,OSTEP,ResType,Resolvent),Resolvents),
	append(Resolvents,Acc1,Acc2),
	ifOption(debugLevel,DL,(DL>3),(write('Consequences so far:'),nl,printClauses(spass,debug,Acc2))),
        modSetproduct(Tail, [OSTEP,NSTN], Acc2, ResType, ResolventSet).
modSetproduct([ML-Clause|Tail], [OSTEP,NSTN], Acc1, ResType, ResolventSet) :-
	findall(Resolvent,maximalResolvent(ML,pos,Clause,NSTN,ResType,Resolvent),Resolvents),
	append(Resolvents,Acc1,Acc2),
	ifOption(debugLevel,DL,(DL>3),(write('Consequences so far:'),nl,printClauses(spass,debug,Acc2))),
        modSetproduct(Tail, [OSTEP,NSTN], Acc2, ResType, ResolventSet),
	!.

structuresToClauses([],Acc1,Acc2) :-
	list_to_set(Acc1,Acc2),
	!.
structuresToClauses([ML-cl(Literals,Char)|SL],Acc1,Acc2) :-
	structuresToClauses(SL,[cl([ML|Literals],Char)|Acc1],Acc2),
	!.

%---------------------------------------------------------------------------
% st_insertClauses(+CLAUSELIST,+ST1,-ST2)
% add the clause structures in CLAUSELIST to the substitution tree ST1
% resulting in a new substitution tree ST2.

st_insertClauses([],[STEP,STEN],[STEP,STEN]).
st_insertClauses([not(L)-C|CL],[STEP1,STEN1],[STEP3,STEN3]) :-
	!,
	ifOption(debugLevel,DL,(DL>9),(write('Inserting '), write(not(L)), nl, write(C), nl)),
	st_insert_term(L,C,STEN1,STEN2),
	st_insertClauses(CL,[STEP1,STEN2],[STEP3,STEN3]),
	!.
st_insertClauses([M-C|CL],[STEP1,STEN1],[STEP3,STEN3]) :-
	ifOption(debugLevel,DL,(DL>9),(write('Inserting '), write(M), nl, write(C), nl)),
	st_insert_term(M,C,STEP1,STEP2),
	st_insertClauses(CL,[STEP2,STEN1],[STEP3,STEN3]),
	!.

%---------------------------------------------------------------------------
% modSetfactor(+CLAUSES,-FACTORS)
% FACTORS is the list of all factors of clauses in CLAUSES
% with respect to order-restricted factoring.

modSetfactor([],Acc,Acc) :-
	!.
modSetfactor([ML-Clause|ClauseList],Acc1,FactorList) :-
        setofOrNil(Factor,maximalFactor(ML,Clause,Factor),FactorList1),
	append(FactorList1,Acc1,Acc2),
	modSetfactor(ClauseList,Acc2,FactorList),
	!.


%---------------------------------------------------------------------------
% maximalResolvent(+CLAUSE1,+MLits1,+CLAUSE2,+MLits2,+RES_TYPE,-CLAUSE3)
% CLAUSE3 is a resolvent of CLAUSE1 and CLAUSE2 such that the resolvent 
% literals are maximal in the respective clauses.

maximalResolvent(MaximalLiteral1,Polarity,cl(Literals1,char(PT1,_VL1,_CRL1,_FL1,_W1)),ST,ResType,Clause) :-
	(ResType == sres ->
	    st_unify(MaximalLiteral1,ST,cl(Literals2,char(PT2,_VL2,_CRL2,_FL2,_W2)))
	;
	    st_unify(MaximalLiteral1,ST,cl(Literals2,char(PT2,[_],_CRL2,_FL2,_W2)))
	),
% Less efficient alternative to the above
%        st_unify(MaximalLiteral1,ST,cl(Literals2,char(PT2,VL2,_CRL2,_FL2,_W2))), 
%	(ResType == tres ->
%	    VL2 \== []
%	;
%	    ResType == sres
%	),
        ifOption(debugLevel,DL,(DL>4),(
                 write('Resolving '), 
		 (Polarity == pos ->
		     printClauses(spass,debug,[cl([MaximalLiteral1|Literals1],char(PT1,2,3,4,5))])
		 ;
		     printClauses(spass,debug,[cl([not(MaximalLiteral1)|Literals1],char(PT1,2,3,4,5))])
		 ),
		 write('with '),
		 (Polarity == pos ->
		     printClauses(spass,debug,[cl([not(MaximalLiteral1)|Literals2],char(PT2,2,3,4,5))])
		 ;
		     printClauses(spass,debug,[cl([MaximalLiteral1|Literals2],char(PT2,2,3,4,5))])
		 ),
		 write('on literal '), write(MaximalLiteral1),nl)),
	modUnion(Literals1,Literals2,Literals),
%	modUnion(PT1,PT2,PT),
	constructProofTerm(PT1,PT2,res,PT),
	ifOption(debugLevel,DL,(DL>4),(write('Derived '), printClauses(spass,debug,[cl(Literals,char(PT,2,3,4,5))]))),
	not(modRemoveTautologies([cl(Literals,char(PT,_X,_Y,_Z,_U))],[])),     % wenn tautologie dann fail.
	computeCharacteristic(cl(Literals),PT,Clause).

constructProofTerm(PT1,PT2,Op,PT3) :-
	(option(proofTerms,yes) ->
	    PT3 = d(Op,PT1,PT2)
	;
	    constructProofTerm1(PT1,PT2,Op,PT3)
	),
	!.

constructProofTerm1([],[PT],fact,d(D,fact,[PT])) :-
	gensym(d,D),
	!.
constructProofTerm1([],d(D1,_,_),fact,d(D,fact,[D1])) :-
	gensym(d,D),
	!.
constructProofTerm1([PT1],[PT2],Op,PT3) :-
	gensym(d,D),
	PT3 = d(D,Op,[PT1,PT2]),
	!.
constructProofTerm1(d(D1,_,_),[PT2],Op,PT3) :-
	gensym(d,D),
	PT3 = d(D,Op,[D1,PT2]),
	!.
constructProofTerm1([PT1],d(D1,_,_),Op,PT3) :-
	gensym(d,D),
	PT3 = d(D,Op,[PT1,D1]),
	!.
constructProofTerm1(d(D1,_,_),d(D2,_,_),Op,PT3) :-
	gensym(d,D),
	PT3 = d(D,Op,[D1,D2]),
	!.
constructProofTerm1(PT1,PT2,_,deriv) :-
	write('Unexpected proof terms '), nl,
	write(PT1), nl,
	write(PT2), nl,
	halt,
	!.

modFindSkolLiterals([],[]) :- !.
modFindSkolLiterals([not(H)|T],L) :-
	modFindSkolLiterals([H],L0),
	!,
	modFindSkolLiterals(T,L1),
	(L0 = [] -> 
	    L = L1;
	    L = [not(H)|L1]),
        !.
	
modFindSkolLiterals([H|T],L) :-
	arg(3,H,F),(var(F);atom(F)),
	!,
	modFindSkolLiterals(T,L),
	!.
modFindSkolLiterals([H|T],[H|L]) :-
	modFindSkolLiterals(T,L),
	!.

	
modNegateAtoms([],[]) :-
	!.
modNegateAtoms([PL|PLL],[not(PL)|PLN]) :-
	modNegateAtoms(PLL,PLN),
	!.

modNegateLiteral(L1,L2) :-
	(L1 = not(L) ->
	    L2 = L
	;
	    L2 = not(L1)
	).

%---------------------------------------------------------------------------
% maximalFactor(+MAXIMALLITERAL,+RESTCLAUSE,+MAXIMALLITERAL'-'-FACTOR)
% FACTOR is a factor of [MAXIMALLITERAL|RESTCLAUSE].

maximalFactor(ML,cl(Literals1,char(PT,C2,C3,C4,C5)),
              ML-cl(Literals2,char(PT2,C2,C3,C4,C5))) :-
	memberPos(L1,Literals1,Pos),
	modUnify(ML,L1),
	modDeletePos(Pos,Literals1,Literals2),
	constructProofTerm([],PT,fact,PT2),
	ifOption(debugLevel,DL,(DL>4),(write('Derived '), printClauses(spass,debug,[cl(Literals2,char(fact(PT),C2,C3,C4,C5))]))).


memberPos(X,[X|_],0).
memberPos(X,[_|L],N) :-
	memberPos(X,L,M),
	N is M+1.

modDeletePos(0,[_X|L],L).
modDeletePos(N,[X|L],[X|LL]) :-
	N > 0,
	M is N-1,
	modDeletePos(M,L,LL).

%---------------------------------------------------------------------------
% modRemoveTautologies(+CLAUSELIST1,-CLAUSELIST2)
% CLAUSELIST2 contains all non-tautological clauses of CLAUSELIST1

modRemoveTautologies([],[]) :-
	!.
modRemoveTautologies([cl(Literals,char(PT,_V,_L,_SF,_W))|CL],CLN) :-
	member(not(L1),Literals),
	member(L2,Literals),
	L1 == L2,
        ifOption(debugLevel,DL,(DL>4),
	         (write('Tautology '), printClauses(spass,debug,[cl(Literals,char(PT,2,3,4,5))]))),
	modRemoveTautologies(CL,CLN).
modRemoveTautologies([ML-cl(Literals,char(PT,_V,_L,_SF,_W))|CL],CLN) :-
	member(not(L1),[ML|Literals]),
	member(L2,[ML|Literals]),
	L1 == L2,
        ifOption(debugLevel,DL,(DL>4),
	         (write('Tautology '), printClauses(spass,debug,[cl(Literals,char(PT,2,3,4,5))]))),
	modRemoveTautologies(CL,CLN).
modRemoveTautologies([C1|CL],[C1|CLN]) :-
	modRemoveTautologies(CL,CLN).

%---------------------------------------------------------------------------
% all_maxlits_all_clauses(+CLAUSELIST1,ECLAUSELIST2)
% each CLAUSE in CLAUSELIST1 will be transformed into the list of all clause 
% structures of the form MAXIMALLITERAL-CLAUSE1 such that MAXIMALLITERAL 
% is a maximal literal in CLAUSE and CLAUSE1 is the result of deleting
% MAXIMALLITERAL from the list of literals in CLAUSE.

all_maxlits_all_clauses([],Acc,Acc).
all_maxlits_all_clauses([KeyClause|List],Acc1,ListMax) :-
	getClause(KeyClause,Clause),
	getClauseLiterals(Clause,LL),
	getClauseChar(Clause,Char),
	find_maximal_literals(Clause,MaxLits),
	associate_maximal_literals(MaxLits,LL,Char,Clauses),
	append(Clauses,Acc1,Acc2),
	all_maxlits_all_clauses(List,Acc2,ListMax),
	!.

getClause(_-Clause,Clause) :-
	!.
getClause(Clause,Clause) :-
	!.

getClauseLiterals(CL,Literals) :- arg(1,CL,Literals).
getClauseChar(CL,Char)         :- arg(2,CL,Char).

%---------------------------------------------------------------------------
% associate_maximal_literals(+MAXIMALLITERALS,+LITERALLIST,+CHARACTERISTIC,
%                            -LIST_OF_CLAUSE_STRUCTURES)
% for each MAXIMALLITERAL in the list MAXIMALLITERALS, MAXIMALLITERAL is 
% deleted from LITERALLIST and a new clause CLAUSE is constructed consisting
% of the remaining literals and CHARACTERISTIC. A clause structure of the 
% form MAXIMALLITERAL-CLAUSE is created. These clause structures are
% collected in LIST_OF_CLAUSE_STRUCTURES.

associate_maximal_literals([],_Clause,_Char,[]).
associate_maximal_literals([ML|MLL],Clause,Char,[ML-NewClause|NewClauses]) :-
	delete_literal_in_clause(Clause,[],ML,RestClause),
	constructClause(RestClause,Char,NewClause),
	ifOption(debugLevel,DL,(DL>8),
	         (write('Maximal literal '),write(ML),write(' in clause '),
		  write(Clause),nl)),
	associate_maximal_literals(MLL,Clause,Char,NewClauses).

constructClause(Literals,Char,cl(Literals,Char)).

%---------------------------------------------------------------------------
% delete_literal_in_clause(+REMAININGLITERALS,+PREFIXLITERALS,+LITERAL,
%                          -LITERALLIST)
% LITERALLIST is the result of removing LITERAL from REMAININGLITERALS
% and appending PREFIXLITERALS.

delete_literal_in_clause([],Start,_,Start).
delete_literal_in_clause([L1|LL],Start,ML,RestClause) :-
	(ML == L1 ->
	    append(Start,LL,RestClause)
	;
	    delete_literal_in_clause(LL,[L1|Start],ML,RestClause)
	).

%---------------------------------------------------------------------------
% find_maximal_literals(+CLAUSE,MAXIMALLITERALS)
% MAXIMALLITERALS will be the list of maximal literals in CLAUSE.

find_maximal_literals(Clause,MLL) :-
	setofOrNil(LL,maximal_literal(Clause,LL),MLL).

%---------------------------------------------------------------------------
% modForwardList(+CLAUSES1,+CLAUSES2,+CLAUSES3,+STAT1,-CLAUSE4,-CLAUSES5,-STAT2)

modForwardList([],SortedClauseList,ProductiveNewClauses,STAT,
               SortedClauseList,ProductiveNewClauses,STAT) :-
        !.
modForwardList([C|CL],SCL1,ProductiveNewClauses,stat(NR1,NS1),
	              SCL3,CLN,stat(NR6,NS6)) :-
	ifOption(debugLevel,DL,(DL>7),(write('Trying to subsume clause: '),printClauses(spass,debug,[C]), write('by queue clauses'),nl)),
	ifOption(debugLevel,DL,(DL>8),(write('ProductiveNewClauses: '),nl,printClauses(spass,debug,ProductiveNewClauses))),
	ifOption(debugLevel,DL,(DL>8),(write('OldClauses: '),nl,printClauses(spass,debug,SCL1))),
	modFindSubsumingClause(SCL1,C,[],Mode,DerivedClauses),
	!,
	NS2 is NS1+1,
	(Mode = subsumResolv ->
	    ifOption(debugLevel,DL,(DL>7),(write('Starting backward subsumption'),nl)),
	    modBackwardSubsumption(SCL1,DerivedClauses,stat(NR1,NS2),
	                           SCL2,stat(NR3,NS3)),
	    ifOption(debugLevel,DL,(DL>7),(write('ended consideration of OldClauses, next looking at ProductiveNewClauses'),nl)),
	    modBackwardSubsumption(ProductiveNewClauses,DerivedClauses,
	                           stat(NR3,NS3),PNC2,stat(NR4,NS4)),
	    ifOption(debugLevel,DL,(DL>7),(write('Ended backward subsumption'),nl)),
            ifOption(debugLevel,DL,(DL>4),(
                     write('Derived '),
                     printClauses(spass,debug,DerivedClauses),
      	             write('by subsumption resolution'),nl)),
            length(DerivedClauses,NDC),
            keymerge(DerivedClauses,PNC2,PNC),
	    NR5 is NR4+NDC
	;
	    NR5 is NR1,
	    NS4 is NS2,
	    SCL2 = SCL1,
	    PNC  = ProductiveNewClauses
	),
	modForwardList(CL,SCL2,PNC,stat(NR5,NS4),SCL3,CLN,stat(NR6,NS6)),
	!.
modForwardList([C|CL],SCL1,ProductiveNewClauses1,stat(NR1,NS1),
	              SCL3,CLN,stat(NR3,NS3)) :-
	ifOption(debugLevel,DL,(DL>7),(write('failed'), nl)),
	modBackwardSubsumption(SCL1,[C],stat(NR1,NS1),SCL2,stat(NR2,NS2)),
	keymerge([C],ProductiveNewClauses1,PNC2),
	length(CL,NCL),
	ifOption(debugLevel,DL,(DL>7),(write('Performed backward subsumption. '), write(NCL), write(' clauses remaining'), nl)),
	modForwardList(CL,SCL2,PNC2,stat(NR2,NS2),SCL3,CLN,stat(NR3,NS3)),
	!.

%---------------------------------------------------------------------------
% simpleReduction(+NEWCLAUSES1,+OLDCLAUSES,+NEWCLAUSES2,+STAT1,
%                 -REMAININGCLAUSES,-STAT2)
% REMAININGCLAUSES is the list of clauses in NEWCLAUSES1 which are not 
% subsumed by clauses in OLDCLAUSES or NEWCLAUSES2. Any clause passing this
% test will be added to NEWCLAUSES2 before an recursive call to simpleReduction.

simpleReduction([],_SL,ProductiveNewClauses,stat(NR1,NS1),
	                 ProductiveNewClauses,stat(NR1,NS1)) :-
        !.
simpleReduction([C|CL],SL1,ProductiveNewClauses,stat(NR1,NS1),
	                 CLN,stat(NR5,NS5)) :-
	ifOption(debugLevel,DL,(DL>7),(write('Trying to subsume clause: '),printClauses(spass,debug,[C]), write('by old clauses'),nl)),
        srFindSubsumingClause(ProductiveNewClauses,C,[],subsumCheck,_DerivedClauses),
	!,
	ifOption(debugLevel,DL,(DL>7),(write('successful'), nl)),
	NS2 is NS1+1,
	simpleReduction(CL,SL1,ProductiveNewClauses,stat(NR1,NS2),
                                 CLN,stat(NR5,NS5)).
simpleReduction([C|CL],SL1,ProductiveNewClauses,stat(NR1,NS1),
	                 CLN,stat(NR5,NS5)) :-
        srFindSubsumingClause(SL1,C,[],subsumCheck,_DerivedClauses),
	!,
	ifOption(debugLevel,DL,(DL>7),(write('successful'), nl)),
	NS2 is NS1+1,
	simpleReduction(CL,SL1,ProductiveNewClauses,stat(NR1,NS2),
                                 CLN,stat(NR5,NS5)).
simpleReduction([C|CL],SL1,ProductiveNewClauses,stat(NR1,NS1),
	                 CLN,stat(NR5,NS5)) :-
	ifOption(debugLevel,DL,(DL>7),(write('failed'), nl)),
        modBackwardSubsumption(ProductiveNewClauses,[C],stat(NR1,NS1),
                               ResultingClauses,stat(NR2,NS2)),
        !,
	simpleReduction(CL,SL1,[C|ResultingClauses],stat(NR2,NS2),
                                 CLN,stat(NR5,NS5)).

srFindSubsumingClause([Clause2|ClauseList],Clause1,DerivedClauses1,
	               subsumCheck,DerivedClauses2) :-
	modSubsumes(Clause2,Clause1,subsumCheck,Clause),
	!,
	modFSCcont(subsumCheck,Clause2,Clause1,Clause,ClauseList,DerivedClauses1,
                   subsumCheck,DerivedClauses2), 
	!.
srFindSubsumingClause([_Clause2|ClauseList],Clause1,DerivedClauses1,
	               Mode,DerivedClauses2) :-
	srFindSubsumingClause(ClauseList,Clause1,DerivedClauses1,
                               Mode,DerivedClauses2),
	!.

%---------------------------------------------------------------------------
% modForwardSubsumption(+CL1,+ST1,+CL2,-ST2,-CL3)
% CL3 will be the list of all clauses in CL1 which are not
% subsumed by clauses in ST1. ST2 will contain those clause in ST1
% which are not subsumed by clauses in CL3.

modForwardSubsumption([],[ST,SL],ProductiveNewClauses,stat(NR1,NS1),
	              [ST,SL],ProductiveNewClauses,stat(NR1,NS1)) :-
        !.
modForwardSubsumption([C|CL],[ST1,SL1],ProductiveNewClauses,stat(NR1,NS1),
	              [ST3,SL3],CLN,stat(NR5,NS5)) :-
	ifOption(debugLevel,DL,(DL>7),(write('Trying to subsume clause: '),printClauses(spass,debug,[C]), write('by old clauses'),nl)),
	ifOption(debugLevel,DL,(DL>8),(write('ProductiveNewClauses: '),nl,printClauses(spass,debug,ProductiveNewClauses))),
	ifOption(debugLevel,DL,(DL>8),(write('OldClauses: '),nl,printClauses(spass,debug,SL1))),
	(
	    modFindSubsumingClause(ProductiveNewClauses,C,[],Mode,DerivedClauses)
	;
	    modFindSubsumingClause(SL1,C,[],Mode,DerivedClauses)
	),
	!,
	NS2 is NS1+1,
	ifOption(debugLevel,DL,(DL>7),(write('succesful with mode '),write(Mode),nl)),
	(Mode = subsumResolv ->
	    length(DerivedClauses,NDC),
	    NR2 is NR1+NDC,
	    ifOption(debugLevel,DL,(DL>7),(write('Starting backward subsumption'),nl)),
	    modBackwardSubsumption(SL1,ST1,DerivedClauses,stat(NR2,NS2),
	                           DerivedClauses1,SL2,ST2,stat(NR3,NS3)),
	    append(DerivedClauses1,DerivedClauses,DerivedClauses2),
	    modBackwardSubsumption(ProductiveNewClauses,DerivedClauses2,
	                           stat(NR3,NS3),
	                           ProductiveNewClauses1,stat(NR4,NS4)),
	    ifOption(debugLevel,DL,(DL>7),(write('Ended backward subsumption'),nl)),
            ifOption(debugLevel,DL,(DL>2),(
                     write('Derived '), nl,
                     printClauses(spass,debug,DerivedClauses),
      	             write('by subsumption resolution'),nl)),
	    append(DerivedClauses2,ProductiveNewClauses1,PNC)
	;
	    NR4 = NR1,
	    NS4 = NS2,
	    SL2 = SL1,
	    ST2 = ST1,
	    PNC = ProductiveNewClauses
	),
	modForwardSubsumption(CL,[ST2,SL2],PNC,stat(NR4,NS4),
                              [ST3,SL3],CLN,stat(NR5,NS5)),
	!.
modForwardSubsumption([C|CL],[ST1,SL1],
	              ProductiveNewClauses0,stat(NR1,NS1),
	              [ST3,SL3],CLN,stat(NR4,NS4)) :-
	ifOption(debugLevel,DL,(DL>7),(write('failed'),nl)),
	modBackwardSubsumption(SL1,ST1,[C],stat(NR1,NS1),
	                       DerivedClauses,SL2,ST2,stat(NR2,NS2)),
%	SL2 = SL1,
%	ST2 = ST1,
	modBackwardSubsumption(ProductiveNewClauses0,[C],stat(NR2,NS2),
	                       ProductiveNewClauses1,stat(NR3,NS3)),
        append(DerivedClauses,ProductiveNewClauses1,ProductiveNewClauses2),
	modForwardSubsumption(CL,[ST2,SL2],
	                      [C|ProductiveNewClauses2],stat(NR3,NS3),
			      [ST3,SL3],CLN,stat(NR4,NS4)),
        !.

%---------------------------------------------------------------------------
% modFindSubsumingClause(+CLAUSELIST,+CLAUSE)
% true iff CLAUSELIST contains a clause subsuming CLAUSE.

modFindSubsumingClause([Clause2|ClauseList],Clause1,DerivedClauses1,
	               Mode2,DerivedClauses2) :-
	modSubsumes(Clause2,Clause1,Mode1,Clause),
	!,
	modFSCcont(Mode1,Clause2,Clause1,Clause,ClauseList,DerivedClauses1,
                   Mode2,DerivedClauses2), 
	!.
modFindSubsumingClause([_Clause2|ClauseList],Clause1,DerivedClauses1,
	               Mode,DerivedClauses2) :-
	modFindSubsumingClause(ClauseList,Clause1,DerivedClauses1,
                               Mode,DerivedClauses2),
        !.
modFindSubsumingClause([],_Clause1,[C|CL],
                       subsumResolv,[C|CL]) :-
	!.

modFSCcont(subsumCheck,Clause2,Clause1,_Clause,_ClauseList,
           _DerivedClauses1,subsumCheck,[]) :-
	ifOption(debugLevel,DL,(DL>7),(
	         write('Subsumed '), 
		 printClauses(spass,debug,[Clause1]),
		 write('by '), printClauses(spass,debug,[Clause2]))),
        !.
modFSCcont(subsumption,Clause2,Clause1,_Clause,_ClauseList,
           _DerivedClauses1,subsumCheck,[]) :-
	ifOption(debugLevel,DL,(DL>7),(
	         write('Subsumed '), 
		 printClauses(spass,debug,[Clause1]),
		 write('by '), printClauses(spass,debug,[Clause2]))),
        !.
modFSCcont(subsumResolv,Clause2,Clause1,Clause,ClauseList,
           DerivedClauses1,Mode,DerivedClauses2) :-
	ifOption(debugLevel,DL,(DL>7),(
                 write('Subsumption resolution reduced '), 
		 printClauses(spass,debug,[Clause1]),
		 write('by '), printClauses(spass,debug,[Clause2]),
	         write('to '), printClauses(spass,debug,[Clause]))),
        modFindSubsumingClause(ClauseList,Clause1,[Clause|DerivedClauses1],
                               Mode,DerivedClauses2),
        !.


%---------------------------------------------------------------------------
% modBackwardSubsumption(+CLAUSELIST1,+CLAUSELIST2,-CLAUSELIST3)
% CLAUSELIST3 will be the list of all clauses in CLAUSELIST1 which are not
% subsumed by clauses in CLAUSELIST2.

modBackwardSubsumption([],_NewClauses,S,[],S) :-
	!.
modBackwardSubsumption([OldClause|OldClauses],NewClauses,stat(NR1,NS1),ResultingClauses,stat(NR4,NS3)) :-
	ifOption(debugLevel,DL,(DL>7),(write('Trying to backward subsume clause: '),printClauses(spass,debug,[OldClause]), write('by new clauses'),nl)),
	modFindSubsumingClause(NewClauses,OldClause,[],Mode,Clauses),
	!,
	ifOption(debugLevel,DL,(DL>7),(write('succesful with mode '),write(Mode),nl)),
	NS2 is NS1+1,
	modBackwardSubsumption(OldClauses,NewClauses,stat(NR1,NS2),
	                       AllClauses,stat(NR3,NS3)),
	(Mode = subsumResolv ->
	    length(Clauses,NC),
	    NR4 is NR3+NC,
	    append(Clauses,AllClauses,ResultingClauses)
	;
	    NR4 = NR3,
	    ResultingClauses = AllClauses
	),
	!.
modBackwardSubsumption([OldClause|OldClauses],NewClauses,S1,[OldClause|AllClauses],S2) :-
	ifOption(debugLevel,DL,(DL>7),(write('failed.'),nl)),
	modBackwardSubsumption(OldClauses,NewClauses,S1,AllClauses,S2),
	!.


modBackwardSubsumption([],[STP,STN],_NewClauses,S,
                       [],[],[STP,STN],S) :-
        !.
modBackwardSubsumption([OldClause|OldClauses],[STP1,STN1],NewClauses,
	               stat(NR1,NS1),
		       DerivedClauses,AllClauses,[STP2,STN2],stat(NR4,NS3)) :-
	modFindSubsumingClause(NewClauses,OldClause,[],Mode,Clauses),
	!,
	NS2 is NS1+1,
	st_deleteClause(OldClause,[STP1,STN1],[STP3,STN3]),
	modBackwardSubsumption(OldClauses,[STP3,STN3],NewClauses,stat(NR1,NS2),
                               DerivedClauses1,AllClauses,[STP2,STN2],stat(NR3,NS3)),
	(Mode = subsumResolv ->
	    length(Clauses,NDC),
	    NR4 is NR3+NDC,
	    append(Clauses,DerivedClauses1,DerivedClauses)
	;
	    NR4 = NR3,
	    DerivedClauses = DerivedClauses1
	),
	!.
modBackwardSubsumption([OldClause|OldClauses],[STP1,STN1],NewClauses,S1,
                       DerivedClauses,[OldClause|AllClauses],[STP2,STN2],S2) :-
	modBackwardSubsumption(OldClauses,[STP1,STN1],NewClauses,S1,
                               DerivedClauses,AllClauses,[STP2,STN2],S2),
	!.

st_deleteClause(cl(Lits,char(PT,_,_,_,_)),[STP1,STN1],[STP2,STN2]) :-
	ifOption(debugLevel,DL,(DL>8),(write('Deleting '), printClauses(spass,debug,[cl(Lits,_)]), write(' from index.'),nl)),
	st_deleteLiterals(Lits,PT,[STP1,STN1],[STP2,STN2]),
	!.

st_deleteLiterals([],_,[STP1,STN1],[STP1,STN1]).
st_deleteLiterals([not(L)|Lits],PT,[STP1,STN1],[STP2,STN2]) :-
	!,
	st_delete_term(L,STN1,STN3,user:equalClauses(PT)),
	st_deleteLiterals(Lits,PT,[STP1,STN3],[STP2,STN2]),
	!.
st_deleteLiterals([L|Lits],PT,[STP1,STN1],[STP2,STN2]) :-
	!,
	st_delete_term(L,STP1,STP3,user:equalClauses(PT)),
	st_deleteLiterals(Lits,PT,[STP3,STN1],[STP2,STN2]),
	!.

user:equalClauses(PT1,cl(_L,char(PT2,_,_,_,_))) :-
	PT1 == PT2.
	
%---------------------------------------------------------------------------
% computeCharacteristic(+CLAUSE1,+CLAUSE2)
% CLAUSE1 is either of the form 
%         cl(POSITIVELITERALS,NEGATIVLITERALS)
%      or cl(POSITIVELITERALS,NEGATIVLITERALS,PROOFTERM).
% CLAUSE2 is of the form cl(LITERALS,char(PROOFTERM,
%                                         LIST_OF_VARIABLES,
%                                         LIST_OF_PREDICATES,
%                                         LIST_OF_SKOLEM_FUNCTIONS,
%                                         WEIGHT))
% where LITERALS is the result of negating all literals in 
% NEGATIVLITERALS and appending POSITIVELITERALS.


computeCharacteristic(cl(Literals),PT,cl(Literals,char([PT],VL,PL,FL,Weight))) :-
	atomic(PT),
	!,
	modExtractPredicates(Literals,PL,NumberOfPredicates),
	modExtractFunctions(Literals,FL,NumberOfFunctions),
	term_variables(Literals,VL),
	Weight is NumberOfPredicates + NumberOfFunctions.
computeCharacteristic(cl(Literals),PT,cl(Literals,char(PT,VL,PL,FL,Weight))) :-
	\+ atomic(PT),
	!,
	modExtractPredicates(Literals,PL,NumberOfPredicates),
	modExtractFunctions(Literals,FL,NumberOfFunctions),
	term_variables(Literals,VL),
	Weight is NumberOfPredicates + NumberOfFunctions.
computeCharacteristic(cl(Pos,Neg),Clause) :-
	gensym(c,PT),
	computeCharacteristic(cl(Pos,Neg,PT),Clause).
computeCharacteristic(cl(Literals),Clause) :-
	gensym(c,PT),
	computeCharacteristic(cl(Literals),PT,Clause).
computeCharacteristic(cl(Pos,Neg,PT),
	              cl(Literals,char([PT],VL,PL,FL,Weight))) :-
        modNegateAtoms(Neg,NegatedAtoms),
	ord_union(Pos,NegatedAtoms,Literals),
	modExtractPredicates(Literals,PL,NumberOfPredicates),
	modExtractFunctions(Literals,FL,NumberOfFunctions),
% SICStus Prolog 3
	term_variables(Literals,VL),
	Weight is NumberOfPredicates + NumberOfFunctions,
	!.

computeCharacteristic([],[]) :-
	!.
computeCharacteristic([C|CL],[CS|CSL]) :-
	computeCharacteristic(C,CS),
	computeCharacteristic(CL,CSL),
	!.

modExtractPredicates([],[],0) :-
	!.
modExtractPredicates([H|T],L0,NumberOfLiterals) :-
	modExtractPredicate(H,L1),
	!,
	modExtractPredicates(T,L,NumberOfLiteralsT),
	NumberOfLiterals is NumberOfLiteralsT + 1,
	ord_union(L,L1,L0),
	!.

modExtractPredicate(L0,L1) :-
	(L0 = not(Literal) ->
	    (modExtractPredicate(Literal,[Name]) ->
		L1 = [Name]
	    ;
		L1 = []
	    )
	;
	    functor(L0,Name,_),
	    L1 = [Name]
	).

modExtractFunctions([],[],0).
modExtractFunctions([H|T],L,NumberOfFunctions) :-
	modExtractFunctions(H,L1),	
	modExtractFunctions(T,L2,NumberOfFunctionsT),
	NumberOfFunctions is NumberOfFunctionsT + 1,
	ord_union(L1,L2,L),
	!.

modExtractFunctions(not(Literal),FL) :-
	modExtractFunctions(Literal,FL),
	!.
modExtractFunctions(A,FL) :-
	A =.. [_|Args],
	modExtractFunctionsArgs(Args,FL).

modExtractFunctionsArgs([],[]) :-
	!.
modExtractFunctionsArgs([T1|TL],FL) :-
	modExtractFunctionsArgs(TL,FL0),
	(var(T1) ->
	    FL = FL0
	;
	    (atomic(T1) ->
		ord_union([T1],FL0,FL)
	    ;
		T1 =.. [F|Args],
		ord_union([F],FL0,FL2),
		modExtractFunctionsArgs(Args,FL1),
		ord_union(FL1,FL2,FL)
	    )
	).

modExtractVars([],[]).
modExtractVars([H|T],L) :-
	modExtractVars1(H,L1),
	modExtractVars(T,L2),
%	modUnion(L1,L2,L).
	ord_union(L2,L1,L),
	!.

modExtractVars1(not(Literal),VL) :-
	modExtractVars1(Literal,VL).
modExtractVars1(A,VL) :-
	A =.. [_|Args],
	modExtractVarsArgs(Args,VL).

modExtractVarsArgs([],[]) :-
	!.
modExtractVarsArgs([T1|TL],FL) :-
	modExtractVarsArgs(TL,FL0),
	(var(T1) ->
	    FL = FL0
	;
	    (atomic(T1) ->
		FL0 = FL
	    ;
		T1 =.. [_|Args],
		modExtractVarsArgs(Args,FL1),
		modUnion(FL0,FL1,FL)
	    )
	).

maximal_literal(cl(Literals,Char),Literal) :-
	member(Literal,Literals),
	modMaximalLiteral(Literal,cl(Literals,Char)).
maximal_literal(_-cl(Literals,Char),Literal) :-
	member(Literal,Literals),
	modMaximalLiteral(Literal,cl(Literals,Char)).

printLoopClauses(false) :-
	!,
	write('no loop found'), nl.
printLoopClauses(true) :-
	!,
	write('loop in true'), nl.
printLoopClauses(CL) :-
	!,
	write('the following clauses form a loop: '), nl,
	printClauses(spass,debug,CL),
	nl.

printClauses(_,_,[]).
printClauses(list,_,[C|CL]) :-
	write('Clause: '),
	write(C),
	nl,
	printClauses(list,_,CL),
	!.
printClauses(ksolver,_,[cl(Literals,char(PT,_,_,_,_))|CL]) :-
	!,
	write([PT]),
	write('   '),
	write(Literals),
	nl,
	printClauses(ksolver,_,CL),
	!.
printClauses(ksolver,_,[cl(L)|CL]) :-
	write('      '), 
	write(L),
	nl,
	printClauses(ksolver,_,CL),
	!.
printClauses(saturate,input,[cl(Literals,char(_PT,_,_,_,_))|CL]) :-
	!,
	splitLiterals(Literals,Pos,Neg),
	write('axiom( ['),
	write(Neg),
	write(' , '),
	write(Pos),
	write(' ]).'),
	nl,
	printClauses(spass,input,CL),
	!.
printClauses(saturate,debug,[cl(Literals,char(PT,_,_,_,_))|CL]) :-
	!,
	splitLiterals(Literals,Pos,Neg),
	write(PT),
	write('   '),
	printLiterals(saturate,Neg),
	write(' -> '),
	printLiterals(saturate,Pos),
	write('.'),
	nl,
	printClauses(saturate,debug,CL),
	!.
printClauses(saturate,output,[cl(Literals,char(_PT,_,_,_,_))|CL]) :-
	!,
	not(not((numbervars(Literals,0,_),
	         splitLiterals(Literals,Pos,Neg),
         	 printLiterals(saturate,Neg),
	         write(' -> '),
	         printLiterals(saturate,Pos)))),
        write('.'),
	nl,
	printClauses(satuarte,output,CL),
	!.
printClauses(otter,_,[cl(Literals,char(_PT,_,_,_,_))|CL]) :-
	!,
	not(not((numbervars(Literals,0,_),
	         printLiterals(otter,Literals)))),
	write('.'),
	nl,
	printClauses(otter,_,CL),
	!.
printClauses(spass,_,[cl(Literals,char(PT,_,_,_,_))|CL]) :-
	!,
	write(PT), 
	write('   '),
	write('clause('),
	not(not((term_variables(Literals,FVL),
 	         numbervars(Literals,0,_),
		 (FVL == [] ->
		     write('or('),
		     printLiterals(spass,Literals),
		     write(')')
		 ;
		     write('forall('),
		     write(FVL),
		     write(','),
		     write('or('),
		     printLiterals(spass,Literals),
		     write('))')
		 )))),
	write(').'),
	nl,
	printClauses(spass,_,CL),
	!.
printClauses(P,M,[cl(L)|CL]) :-
	printClauses(P,M,[cl(L,char(1,2,3,4,5))|CL]).
printClauses(P,M,[cl(L1,L2)|CL]) :-
	modNegateLiterals(L2,NL2),
	append(L1,NL2,L),
	printClauses(P,M,[cl(L,char(1,2,3,4,5))|CL]).
printClauses(P,M,[cl(L1,L2,N)|CL]) :-
	atomic(N),
	modNegateLiterals(L2,NL2),
	append(L1,NL2,L),
	printClauses(P,M,[cl(L,char(N,2,3,4,5))|CL]).
printClauses(P,M,[_W-C|CL]) :-
	printClauses(P,M,[C|CL]).

printLiterals(_,[]).
printLiterals(saturate,[L1]) :-
	write(L1).
printLiterals(saturate,[L1,L2|LL]) :-
	write(L1),
	write(','),
	printLiterals(saturate,[L2|LL]),
	!.
printLiterals(otter,[L1]) :-
	printLiteral(otter,L1).
printLiterals(otter,[L1,L2|LL]) :-
	printLiteral(otter,L1),
	write(' | '),
	printLiterals(otter,[L2|LL]),
	!.
printLiterals(spass,[L1]) :-
	printLiteral(spass,L1),
	!.
printLiterals(spass,[L1,L2|LL]) :-
	printLiteral(spass,L1),
	write(', '),
	printLiterals(spass,[L2|LL]),
	!.

printLiteral(otter,not(A)) :-
	!,
	write('-'),
	printLiteral(otter,A).
printLiteral(otter,A) :-
	write(A),
	!.
printLiteral(spass,not(A)) :-
	!,
	write('not('),
	printLiteral(spass,A),
	write(')').
printLiteral(spass,A) :-
	write(A),
	!.

%---------------------------------------------------------------------------
% modNegateLiterals(+LIST1,-LIST2)
% LIST2 is the list of negation normal forms of the negation of the literals
% in LIST1.

modNegateLiterals([],[]).
modNegateLiterals([not(L)|LL],[L|NL]) :-
	!,
	modNegateLiterals(LL,NL).
modNegateLiterals([A|LL],[not(A)|NL]) :-
	!,
	modNegateLiterals(LL,NL).

%---------------------------------------------------------------------------
% splitLiterals(+LITERAL_LIST,-POSITIVE_LITERALS,-NEGATIVE_LITERALS)
% POSITIVE_LITERALS contains all positive literals in LITERAL_LIST
% NEGATIVE_LITERALS contains all negative literals in LITERAL_LIST
% The literals in the result lists are in standard notation.

splitLiterals([],[],[]).
splitLiterals([not(A)|LL],PL,[A|NL]) :-
	!,
	splitLiterals(LL,PL,NL).
splitLiterals([A|LL],[A|PL],NL) :-
	!,
	splitLiterals(LL,PL,NL).

%---------------------------------------------------------------------------
% modJoin(+LIST1,+LIST2,-LIST3)
% If LIST1 and LIST2 are not syntactically equal, LIST3 is the result of 
% appending LIST2 to LIST1. Otherwise, LIST3 is LIST1.

modJoin(CL1,CL2,CL1) :-
	CL1 == CL2,
	!.
modJoin(CL1,CL2,CL3) :-
	append(CL1,CL2,CL3).

%---------------------------------------------------------------------------
% modMaximalLiteral(+LITERAL,+CLAUSE)
% true iff LITERAL covers all literals in CLAUSE.

modMaximalLiteral(L1,cl(Literals,_Char)) :-
	literalToTerm(L1,LT),
	modMaximalLiteralList(Literals,LT),
	!.

modMaximalLiteralList([],_A).
modMaximalLiteralList([H|T],A) :-
	literalToTerm(H,L),
	\+greater(L,A),
	modMaximalLiteralList(T,A).

literalToTerm(not(H),L) :-
	literalToTerm(H,L),
	!.
literalToTerm(A,A) :-
	!.

%modLiteralCovers(A,B) :-
%	functor(A,_,AA),
%	functor(B,_,AB),
%	modLiteralCovers(A,B,AA,AB).
%
%modLiteralCovers(_A,_B,_,0) :- !.
%modLiteralCovers(A,B,AA,AB) :-
%	arg(AB,B,SB),
%	modLiteralCoversSB(A,SB,AA),
%	AB1 is AB-1,
%	modLiteralCovers(A,B,AA,AB1).
%	
%modLiteralCoversSB(A,SB,AA) :-
%	AA >= 1,
%	arg(AA,A,SA),
%%	(do_CoversTermtest(SA,SB) ->
%	(modCovers1(SA,SB) ->            %pmb
%	    true
%	;
%	    AA1 is AA-1,
%	    modLiteralCoversSB(A,SB,AA1)
%	).
%
%do_CoversTermtest(T1,T2) :-
%	modCovers(T1,T2).
%do_CoversTermtest(T1,T2) :-
%	modSubterm(T2,T1).
%
%
%modCovers(X,Y) :-  
%	var(X),!,
%	X == Y.
%modCovers(Clause1,Clause2) :-
%%       t = f(t1,...,tn) , s is var and s = ti for some i ,i=1...n.
%	var(Clause2),
%	!,
%	functor(Clause1,_,A),
%	var_equal_subterm(A,Clause1,Clause2).
%modCovers(Clause1,Clause2) :-
%%       t= f(t1,..,tn)  s= g(t1,..tm)  , n>=m>=0
%	functor(Clause1,_,A1),
%	functor(Clause2,_,A2),
%	A1 >= A2,
%	var_equal_subterms(A2,Clause1,Clause2),
%	!.
%modCovers(Clause1,Clause2) :-
%%       t=s
%	nonvar(Clause2),
%	nonvar(Clause1),
%	not(not(Clause1 = Clause2)).
%
%
%var_equal_subterm(A,Term,Var) :-
%	A >= 1,
%	arg(A,Term,V),
%	(Var == V -> 
%	    true 
%	;
%	    A1 is A-1,
%	    var_equal_subterm(A1,Term,Var)
%	).
%
%var_equal_subterms(0,_Term0,_Term1) :- !.
%var_equal_subterms(A,Term0,Term1) :-
%	A >= 1,
%	arg(A,Term0,V0),
%	arg(A,Term1,V1),
%	(V0 == V1 -> 
%	    true 
%	;
%	    A1 is A-1,
%	    var_equal_subterms(A1,Term0,Term1)
%	).
% 
%modSubterm(A1,A2) :-
%        A1 == A2,
%        !.
%modSubterm(A1,A2) :-
%	nonvar(A2),
%	functor(A2,_,A),
%	modSubterm(A1,A2,A).
%
%modSubterm(A1,A2,A) :-
%	A >= 1,
%	arg(A,A2,SA2),	
%	(modSubterm(A1,SA2) ->
%	    true
%	;
%	    AM1 is A - 1,
%	    modSubterm(A1,A2,AM1)
%	).
%
%modLiteralCovers1(A,B) :-
%	A =.. LitA,
%	B =.. LitB,
%	!,
%	modLiteralCovers2(LitA,LitB).
%modLiteralCovers2(_,[]) :- !.
%modLiteralCovers2(LitA,[Lit|Lits]) :-
%	modLiteralCovers3(LitA,Lit),
%	!,
%	modLiteralCovers2(LitA,Lits).
%
%modLiteralCovers3([LitA|Lits],LitB) :-
%	(modCovers1(LitA,LitB) ->
%	    true;
%	    modLiteralCovers3(Lits,LitB)).
%
%modCovers1(X,Y) :-  
%	var(X),!,
%	X == Y.
%modCovers1(Clause1,Clause2) :-
%%       t = f(t1,...,tn) , s is var and s = ti for some i ,i=1...n.  motel-case : X = f(Y)
%	var(Clause2),
%	!,
%	arg(1,Clause1,V),
%	Clause2 == V.
%modCovers1(Clause1,_Clause2) :-
%%       t= f(t1,..,tn)  s= g(t1,..tm)  , n>=m>=0   motel-case : f(X) = _
%	arg(1,Clause1,_),
%	!.
%modCovers1(Clause1,Clause2) :-
%	atom(Clause1),atom(Clause2),!.
%

%---------------------------------------------------------------------------
% modUnify(+Term1,+Term2)
% succeeds iff Term1 and Term2 unify (with occurs check) and unifies the
% terms at the same time.

modUnify(X, Y) :-
        var(X),var(Y),!,
        X = Y.        
modUnify(X, Y) :-
        var(X),!,
        occurs_check(Y, X),  
        X = Y.
modUnify(X, Y) :-
        var(Y),!,
        occurs_check(X, Y),
        X = Y.
modUnify(X, Y) :-
        atomic(X),!,
        X = Y.
modUnify(X, Y) :-
        functor(X, F, N),
        functor(Y, F, N),
        modUnify(N, X, Y).
        
modUnify(0, _, _) :- !.
modUnify(N, X, Y) :-
        arg(N, X, Xn),arg(N, Y, Yn),
        modUnify(Xn, Yn),
        M is N-1,
        !,
        modUnify(M, X, Y).

occurs_check(Term, Var) :-
        var(Term),!,
        Term \== Var.
occurs_check(Term, Var) :-
        functor(Term, _, Arity),
        occurs_check(Arity, Term, Var).

occurs_check(0, _, _) :- !.
occurs_check(N, Term, Var) :-
        arg(N, Term, Arg),
        occurs_check(Arg, Var),
        M is N-1,
        !,
        occurs_check(M, Term, Var).

%---------------------------------------------------------------------------
% modUnion(+SET1,+SET2,-SET3)
% SET3 is the set union of SET1 and SET2.

modUnion(A,[],A) :-
	!.
modUnion([],A,A) :-
	!.
modUnion([],[],[]).
modUnion([X|R],[Y|R1],Z):-
	modUnion1([X|R],[Y|R1],Z),
	!.
modUnion([X|R],Y,Z) :-
	modUnion1([X|R],[Y],Z),
	!.
modUnion([X],Y,Z) :-
	modUnion1([X],[Y],Z),
	!.
modUnion(X,[Y],Z) :-
	modUnion1([X],[Y],Z),
	!.
modUnion(X,[Y|R],Z) :-
	modUnion1([X],[Y|R],Z),
	!.
modUnion(X,Y,Z) :-
	modUnion1([X],[Y],Z),
	!.
modUnion1([], Union, Union).
modUnion1([Element|Elements], Set, Union) :-
	memberVarList(Element, Set),
	!,
	modUnion1(Elements, Set, Union).
modUnion1([Element|Elements], Set, [Element|Union]) :-
	modUnion1(Elements, Set, Union),
	!.

%---------------------------------------------------------------------------
% modVarEqualSet(+LIST1,+LIST2)
% succeeds iff LIST1 and LIST2 contain the same elements (up to syntactical
% equality).

modVarEqualSet(L1,L2) :-
	modVarSubset(L1,L2),
	modVarSubset(L2,L1),
	!.

%---------------------------------------------------------------------------
% modVarSubset(+LIST1,+LIST2)
% succeeeds iff every element of LIST1 is syntactically identical to an
% element of LIST2.

modVarSubset([],_) :- !.
modVarSubset([H1|T],List) :-
	memberVarList(H1,List),
	modVarSubset(T,List),
	!.

%---------------------------------------------------------------------------
% modSubset(+LIST1,+LIST2)
% succeeds iff every element of LIST1 is the variant of an element in 
% LIST2 (does not unify any of the elements of LIST1 with elements of LIST2!).

modSubset([], _) :- !.
modSubset([B|C], A) :-
	member(L,A),
        not(not(modVariant(B,L))),
	!,
        modSubset(C, A), 
	!.


%---------------------------------------------------------------------------
% modVariant(+A,+B)
% succeeds iff term A is a variant of term B (and unifies the terms at the
% same time.

modVariant(A,B) :-
	var(A),
	var(B),
	!,
	A = B.
modVariant(A,B) :-
	var(A),
	nonvar(B),
	!,
	fail.
modVariant(A,B) :-
	nonvar(A),
	var(B),
	!,
	fail.
modVariant(A,B) :-
	atomic(A),
	atomic(B),
	!,
	A == B.
modVariant(A,B) :-
	A =.. [F|TLA],
	B =.. [F|TLB],
	!,
	modVariantList(TLA,TLB).
modVariantList([],[]) :-
	!.
modVariantList([A|TLA],[B|TLB]) :-
	modVariant(A,B),
	modVariantList(TLA,TLB).

%---------------------------------------------------------------------------
% modMember(+X,+LIST)
% succeeds iff the variable X occurs in LIST

modMember(X,Y) :-
        nonvar(Y),
        X==Y.
modMember(_X,Y) :-
        var(Y),
        !,fail.
modMember(X, [Y|_]    ) :-
        nonvar(Y),
        X=Y.
modMember(X, [_,Y|_]  ) :-
        nonvar(Y),
        X=Y.
modMember(X, [_,_,Y|_]) :-
        nonvar(Y),
        X==Y.
modMember(X, [_,_,_|L]) :-
        modMember(X, L).

%---------------------------------------------------------------------------
% memberVarList(+X,+LIST)
% succeeds iff X is identical to some elmenet of LIST.

memberVarList(X,[Y|_]) :-
	X == Y,!.
memberVarList(X, [_|T]    ) :-
	memberVarList(X,T).

%---------------------------------------------------------------------------
% genWaitingPred(+L,+A)
% A is a `waiting for L' atom

genWaitingPred(L,A1) :-
	modExtractPredicate(L,[P]),
	atom_chars(P,ACP),
	(L = not(_) ->
	    atom_chars(A1,[119,95,110,111,116,95|ACP])
	;
	    atom_chars(A1,[119,95|ACP])
	).

%---------------------------------------------------------------------------
% genAccPred(+MODALOPERATOR,+AGENT,+CHAR,-RELATION)
% generates a predicate symbol uniquely associated with the tuple
% (MODALOPERATOR,AGENT,CHAR).

genAccPred(M,A,Char,R) :-
	symbol_chars(M,ACM),
	symbol_chars(A,ACA),
	symbol_chars(Char,[AC]),
	append(ACM,[95|ACA],AC0),
	atom_chars(R,[AC,95|AC0]),
	!.

%---------------------------------------------------------------------------
% genModalPred(+MODALOPERATOR,+AGENT,+LITERAL,-ATOM)
% generates a atom ATOM with a predicate symbol uniquely associated with
% the tuple (MODALOPERATOR,AGENT,LITERAL).

genModalPred(M,A,L,A1) :-
	symbol_chars(M,ACM),
	symbol_chars(A,ACA),
	modExtractPredicate(L,[P]),
	atom_chars(P,ACP),
	append(ACM,[95|ACA],AC0),
	(L = not(_) ->
	    append(AC0,[95,110,95|ACP],AC1)
	;
	    append(AC0,[95|ACP],AC1)
	),
	atom_chars(A1,[109,95|AC1]),
	!.

%---------------------------------------------------------------------------
% symbol_chars(+SYMBOL,-CHARS)
% returns the list CHARS of characters of which the symbol SYMBOL consists
% (where SYMBOL is either an atom or a number).

symbol_chars(S,C) :-
	(atom(S) ->
	    atom_chars(S,C)
	;
	    (number(S) ->
		number_chars(S,C)
	    ;
		write('WARNING: '),
		write(S),
		write(' is not a legal identifier'),
		nl,
		!,
		fail
	    )
	).

%---------------------------------------------------------------------------
% generateFilename(+SUFFIX,+VN,+RN,+AN,+ON,+GR,+RANDOM,-FILE)
% generates the file name FILE containing the random formula specified by
% the parameters VN, RN, AN, ON, GR, RANDOM located in directory BASE.
	
generateFilename(SUFFIX,VN,RN,AN,ON,GR,RandomNumber,File) :-
	getProblemBase(BASE),
	getProblemSeries(SN),
	generateAtom([BASE,'pltl-',VN,'-',RN,'-',SN,'-',ON,'-',GR,'/pltl-',VN,'-',RN,'-',AN,'-',ON,'-',GR,'-',RandomNumber,SUFFIX],[],File).

generateAtom([],AtomString,Atom) :-
	atom_chars(Atom,AtomString).
generateAtom([A|L],AtomString1,Atom) :-
	(number(A) ->
	    number_chars(A,AC)
	;
	    (atomic(A) ->
		atom_chars(A,AC)
	    ;
		AC = A
	    )
	),
	append(AtomString1,AC,AtomString2),
	generateAtom(L,AtomString2,Atom).

getProblemBase('/users/lect/ullrich/provers/experiments/pltl/samples/') :-
	!.
getProblemSeries(SN) :-
	testseries(SN),
	!.
getProblemSeries(x) :-
	!.

%---------------------------------------------------------------------------
% gensym(+PREFIX,-ATOM)
% returns a new ATOM consisting of the characters of PREFIX followd
% by a natural number.

gensym(Prefix, V) :-
        var(V),
        atomic(Prefix),
        (   retract(gensym_counter(Prefix, M))
        ;   M = 0

        ),
        N is M+1,
        asserta(gensym_counter(Prefix, N)),
        name(Prefix,P1),
        name(N,N1),
        append(P1,N1,V1),
        name(V,V1),
   !.

%---------------------------------------------------------------------------
% getRuntime(-RT)
% RT is the runtime used so far by the process

getRuntime(RT) :-
	statistics(runtime,[RT|_]).

%---------------------------------------------------------------------------
% setofOrNil(+VAR,+GOAL,-LIST)
% returns in LIST all the VAR that satisfy GOAL

setofOrNil(A,B,C) :-
        (setof(A,B,C) -> true ; C = []).

%---------------------------------------------------------------------------
% not(+GOAL)
% succeeds iff Goal fails

not(Goal) :- 
	call(\+ Goal).

%--------------------------------------------------------------------------- 
% setOption(+OPTION,+VAL)
% set option OPTION to value VAL.

setOption(Option,Set) :-
        retractall(option(Option,_)),
        asserta(option(Option,Set)).

%---------------------------------------------------------------------------
% ifOption(+OPTION,+VAL,+GOAL)
% executes GOAL if the current value of OPTION is VAL otherwise
% the predicate suceeds without executing GOAL.

ifOption(Option,Set,Goal) :-
        (option(Option,Set) ->
	    call(Goal)
	;
	    true
	),
	!.

ifOption(Option,Value,Test,Goal) :-
	((option(Option,Value),
	  call(Test)) ->
	    call(Goal)
	;
	    true
	),
	!.

%---------------------------------------------------------------------------
% retractall(+ENV,+PRED/+ARITY)
% removes all Prolog clauses with predicate PRED of arity ARITY in 
% environment ENV

retractall(Env,Pred/Arity) :-
        constructHead(Env,Pred/Arity,Head),
        retractall(Head).

%---------------------------------------------------------------------------
% setModalLogic(+MODALOPERATOR,+AGENT,+LOGIC)
% assserts that the modality (MODALOPERATOR,AGENT) has the properties 
% specified by LOGIC. LOGIC is a list consisting of a subset of `k',`d',`b',
% `t', `4' and `5'.

setModalLogic(M,A,L) :-
	asserta(modalLogic(M,A,L)).
setModallogic(M,A,L) :-
	setModalLogic(M,A,L).
setML(M,A,L) :-
	setModalLogic(M,A,L).

start_message :-
	write('TRP Version 1.4'), nl,
        write('Copyright (c) 2001-2008 Ullrich Hustadt, University of Liverpool.'), nl,
	write('Copy policy: GPL-2 (see www.gnu.org)'), nl,
	write('Implemented in SICStus Prolog (Version 3.12.x)'), nl.

help_message :-
	write('For help on TRP, use ?- help.'), nl, nl.

start_program :-
	start_message,
	current_prolog_flag(argv,Args),
	(Args == [] ->
	    !,
	    help_message
	;
	    last(Args,File),
	    (File == '--' ->
		!,
		help_message
	    ;
		testFile(File),
		halt
	    )
	).

user_help :-
	format("TRP Version 1.4~n~n",[]),
	format("INPUT SYNTAX:~n",[]),
	format("FORMULA             ::= and(LIST_OF_SNFCLAUSES)~n",[]),
	format("LIST_OF_SNFCLAUSES  ::= [] | [SNFCLAUSE,...,SNFCLAUSE]~n",[]),
	format("SNFCLAUSE           ::= always(TEMPORALCLAUSE) | always(MODALCLAUSE) | PROPOSITIONALCLAUSE~n",[]),
	format("TEMPORALCLAUSE      ::= PROPOSITIONALCLAUSE | GLOBALCLAUSE | SOMETIMECLAUSE~n",[]),
        format("PROPOSITIONALCLAUSE ::= or([LITERAL,...,LITERAL])~n",[]),
        format("GLOBALCLAUSE        ::= or([LITERAL,...,LITERAL,next(LITERAL),...,next(LITERAL)])~n",[]),
	format("SOMETIMECLAUSE      ::= or([LITERAL,...,LITERAL,sometime(LITERAL)])~n",[]),
	format("MODALCLAUSE         ::= or([LITERAL,...,LITERAL,MODALLITERAL,...,MODALLITERAL])~n",[]),
	format("LITERAL             ::= not(IDENTIFIER) | IDENTIFIER~n",[]),
	format("MODALLITERAL        ::= box(MODALOPERATOR,AGENT,LITERAL) | not(box(MODALOPERATOR,AGENT,LITERAL))~n",[]),
	format("MODALOPERATOR       ::= IDENTIFIER~n",[]),
	format("AGENT               ::= IDENTIFIER~n",[]),
	format("IDENTIFIER          ::= [a-z0-0]+~n",[]),
	format("LIST_OF_AXIOMIDS    ::= [] | [AXIOMID,...,AXIOMID]~n",[]),
	format("AXIOMID             ::= k | d | t | b | 4 | 5~n",[]),
	format("EXAMPLE OF A FORMULA:~n",[]),
	format("and([or([p0,not(p1)]),~n",[]),
        format("     always(or([p1,not(p0)])),~n",[]),
        format("     always(or([not(p0), next(not p4), next(p1)])),~n",[]),
        format("     always(or([not(p1), sometime(p2)])),~n",[]),
	format("     always(or([not(p2), box(know,a1,p4)]))])~n~n",[]),
	format("USAGE:~n",[]),
	format("- testFile(FILENAME).~n",[]),
	format("  FILENAME is the name of a file containg a FORMULA followed by `.`'~n",[]),
	format("  FORMULA will be tested for satisfiability, the output will state~n",[]),
	format("  whether FORMULA is satisfiabile and how many inference steps have~n",[]),
        format("  been performed.~n",[]),
	format("- testFormula(FORMULA).~n",[]),
	format("  FORMULA will be tested for satisfiability, the output will state~n",[]),
	format("  whether FORMULA is satisfiabile and how many inference steps have~n",[]),
        format("  been performed.~n",[]),
	format("- setModalLogic(MODALOPERATOR,AGENT,LIST_OF_AXIOMIDS).~n",[]),
	format("  asserts that MODALOPERATOR for AGENT satifies the modal axiomschemata~n",[]),
	format("  listed in LIST_OF_AXIOMIDS, e.g. if you want that the modal operator~n",[]),
	format("  `know' for agent `a1' is a KT45 modality, then use~n",[]),
	format("         setModalLogic(know,a1,[k,t,4,5]).~n",[]),
	format("  If you want that `believe' is a KD45 modality for agent `a2', then use~n",[]),
	format("         setModalLogic(believe,a2,[k,d,4,5]).~n",[]),
	format("- setOption(debugLevel,NUMBER).~n",[]),
	format("  where NUMBER is a number between 0 and 9.~n",[]),
	format("  Higher values increase the verbosity of the procedures testFile and~n",[]),
	format("  testFormula during the derivation,e.g.~n",[]),
	format("         setOption(debugLevel,3).~n",[]),
	format("  shows you (almost all) the inference steps performed during the proof search.~n",[]),
	format("- halt.",[]),
	format("  exiting TRP (and SICStus Prolog).~n",[]),
	!.

:- setOption(proofTerms,no).
%:- initialization((reconsult('/var/tmp/tempRes.pl'),halt)).
:- initialization(start_program).
:- save_program('TRP').

%append([], List, List).
%append([Head|Tail], List, [Head|Rest]) :- 
%        append(Tail, List, Rest).

